#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import os
import warnings
from os import getcwd, listdir
from os.path import join, dirname, basename, splitext

import click
import numpy as np
import tifffile.tifffile as tiff

from tardis_em_analysis.analysis import analyse_filaments_list
from tardis_em.utils.predictor import GeneralPredictor
from tardis_em._version import version
from tardis_em import format_choices
from tardis_em.utils.dataset import find_filtered_files

warnings.simplefilter("ignore", UserWarning)


@click.command()
@click.option(
    "-dir",
    "--path",
    default=getcwd(),
    type=str,
    help="Directory with images for prediction with CNN model.",
    show_default=True,
)
@click.option(
    "-ms",
    "--mask",
    default=False,
    type=bool,
    help="Define if you input tomograms images or binary mask with pre segmented microtubules.",
    show_default=True,
)
@click.option(
    "-cch",
    "--cnn_checkpoint",
    default=None,
    type=str,
    help="Optional list of pre-trained CNN weights",
    show_default=True,
)
@click.option(
    "-dch",
    "--dist_checkpoint",
    default=None,
    type=str,
    help="Optional list of pre-trained DIST weights",
    show_default=True,
)
@click.option(
    "-mv",
    "--model_version",
    default=None,
    type=int,
    help="Optional version of the model from 1 to inf.",
    show_default=True,
)
@click.option(
    "-cnn",
    "--convolution_nn",
    default="fnet_attn",
    type=str,
    help="Select CNN used for semantic segmentation.",
    show_default=True,
)
@click.option(
    "-out",
    "--output_format",
    default="tif_csv",
    type=click.Choice(format_choices),
    help="Type of output files. The First optional output file is the binary mask "
    "which can be of type None [no output], am [Amira], mrc or tif. "
    "Second output is instance segmentation of objects, which can be "
    "output as amSG [Amira], mrcM [mrc mask], tifM [tif mask], "
    "csv coordinate file [ID, X, Y, Z] or None [no instance prediction].",
    show_default=True,
)
@click.option(
    "-ps",
    "--patch_size",
    default=128,
    type=int,
    help="Size of image patch used for prediction. This will break "
    "the tomogram volumes into 3D patches where each patch will be "
    "separately predicted and then stitched back together "
    "with 25% overlap.",
    show_default=True,
)
@click.option(
    "-rt",
    "--rotate",
    default=True,
    type=bool,
    help="If True, during CNN prediction image is rotate 4x by 90 degrees."
    "This will increase prediction time 4x. However may lead to more cleaner"
    "output.",
    show_default=True,
)
@click.option(
    "-ct",
    "--cnn_threshold",
    default="0.1",
    type=str,
    help="Threshold used for CNN prediction.",
    show_default=True,
)
@click.option(
    "-dt",
    "--dist_threshold",
    default=0.5,
    type=float,
    help="Threshold used for instance prediction.",
    show_default=True,
)
@click.option(
    "-pv",
    "--points_in_patch",
    default=900,
    type=int,
    help="Size of the cropped point cloud, given as a max. number of points "
    "per crop. This will break generated from the binary mask "
    "point cloud into smaller patches with overlap.",
    show_default=True,
)
@click.option(
    "-fl",
    "--filter_by_length",
    default=100,
    type=int,
    help="Filtering parameters for microtubules, defining maximum microtubule "
    "length in angstrom. All filaments shorter then this length "
    "will be deleted.",
    show_default=True,
)
@click.option(
    "-cs",
    "--connect_splines",
    default=25,
    type=int,
    help="To address the issue where microtubules are mistakenly "
    "identified as two different filaments, we use a filtering technique. "
    "This involves identifying the direction each filament end points towards and then "
    "linking any filaments that are facing the same direction and are within "
    "a certain distance from each other, measured in angstroms. This distance threshold "
    "determines how far apart two microtubules can be, while still being considered "
    "as a single unit if they are oriented in the same direction.",
    show_default=True,
)
@click.option(
    "-cc",
    "--connect_cylinder",
    default=12,
    type=int,
    help="To minimize false positives when linking microtubules, we limit "
    "the search area to a cylindrical radius specified in angstroms. "
    "For each spline, we find the direction the filament end is pointing in "
    "and look for another filament that is oriented in the same direction. "
    "The ends of these filaments must be located within this cylinder "
    "to be considered connected.",
    show_default=True,
)
@click.option(
    "-dv",
    "--device",
    default=0,
    type=str,
    help="Define which device to use for training: "
    "gpu: Use ID 0 GPU"
    "cpu: Usa CPU"
    "mps: Apple silicon (experimental)"
    "0-9 - specified GPU device id to use",
    show_default=True,
)
@click.option(
    "-db",
    "--debug",
    default=False,
    type=bool,
    help="If True, save the output from each step for debugging.",
    show_default=True,
)
@click.option(
    "-continue",
    "--continue_b",
    default=False,
    type=bool,
    help="If True, continue from the last tomogram that was successfully predicted.",
    show_default=True,
)
@click.option(
    "-test_click",
    "--test_click",
    default=False,
    hidden=True,
)
@click.version_option(version=version)
def main(
    path: str,
    mask: bool,
    cnn_checkpoint: str,
    dist_checkpoint: str,
    model_version: int,
    output_format: str,
    patch_size: int,
    rotate: bool,
    convolution_nn: str,
    cnn_threshold: str,
    dist_threshold: float,
    points_in_patch: int,
    filter_by_length: int,
    connect_splines: int,
    connect_cylinder: int,
    device: str,
    debug: bool,
    continue_b: bool,
    test_click=False,
):
    """
    MAIN MODULE FOR PREDICTION MT WITH TARDIS-EM
    """
    cleanup_list = []

    if output_format.split("_")[1] == "None":
        instances = False
    else:
        instances = True

    checkpoint = [cnn_checkpoint, dist_checkpoint]

    nd2_files = [i for i in listdir(path) if i.endswith(".nd2")]
    if len(nd2_files) > 0:
        from tardis_em.utils.load_data import load_nd2_file
        import tifffile.tifffile as save_tiff

        for i in nd2_files:
            image, _ = load_nd2_file(join(path, i))

            for j in range(image.shape[0]):
                name_file = join(path, i[:-4]) + f"_{j}.tiff"
                cleanup_list.append(name_file)

                save_tiff.imwrite(name_file, image[j, ...])

    predictor = GeneralPredictor(
        predict="Microtubule_tirf",
        dir_s=path,
        binary_mask=mask,
        correct_px=None,
        normalize_px=1.0,
        convolution_nn=convolution_nn,
        checkpoint=checkpoint,
        model_version=model_version,
        output_format=output_format,
        patch_size=patch_size,
        cnn_threshold=cnn_threshold,
        dist_threshold=dist_threshold,
        points_in_patch=points_in_patch,
        predict_with_rotation=rotate,
        amira_prefix=None,
        filter_by_length=filter_by_length,
        connect_splines=connect_splines,
        connect_cylinder=connect_cylinder,
        amira_compare_distance=None,
        amira_inter_probability=None,
        instances=instances,
        device_s=str(device),
        debug=debug,
        continue_b=continue_b,
    )

    if not test_click:
        predictor()

        if len(nd2_files) > 0:
            # Cleanup move tiffs to the Predictions directory
            images = []
            for i in cleanup_list:
                f_name = dirname(i)
                images.append(tiff.imread(i))

                os.rename(
                    i, join(f_name, "Predictions", splitext(basename(i))[0]) + ".tif"
                )

            # Analyze length, average intensity along the spline,
            name_s = find_filtered_files(
                join(dirname(path), "Predictions"), prefix="instances_filter"
            )
            data = []
            for d in name_s:
                d = np.genfromtxt(d, delimiter=",")

                if str(d[0, 0]) == "nan":
                    d = d[1:, :]
                data.append(d)

            analyse_filaments_list(
                data=data,
                names_l=name_s,
                path=join(path, "Predictions"),
                images=images,
                px=None,
            )


if __name__ == "__main__":
    main()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import random
import shutil
from glob import glob
from os import listdir, mkdir
from os.path import isdir, join
from shutil import copyfile, copytree, move
from typing import Optional

from tardis_em.utils.errors import TardisError


def find_filtered_files(directory, prefix="instances_filter", format_b="csv"):
    """
    Finds and retrieves a list of files in the specified directory based on a given
    prefix and format(s). This function can search for files matching the prefix
    along with a single format or multiple formats, returning a list of filtered
    file paths.

    :param directory: The directory path in which the search will be performed.
    :type directory: str
    :param prefix: The prefix that filtered files should match.
        Defaults to 'instances_filter'.
    :type prefix: str, optional
    :param format_b: The format or formats (as string or a list/tuple of strings)
        to search for matching files. Defaults to 'csv'.
    :type format_b: str or list or tuple, optional

    :return: A list of file paths that match the specified prefix and format(s) in
        the directory.
    :rtype: list
    """
    if prefix == "":
        extension = "*"
    else:
        extension = "*_"

    if not isinstance(format_b, list) or not isinstance(format_b, tuple):
        search_pattern = join(directory, f"{extension}{prefix}.{format_b}")

        filtered_files = glob(search_pattern)
    else:
        filtered_files = []
        for i in format_b:
            search_pattern = join(directory, f"{extension}{prefix}.{format_b}")

            filtered_files += glob(search_pattern)

    return filtered_files


def move_train_dataset(
    dir_s: str, coord_format: tuple, with_img: bool, img_format: Optional[tuple] = None
):
    """
    Moves and organizes a training dataset by placing coordinate files and optionally
    image files into appropriate subdirectories.

    :param dir_s: The directory containing dataset files to be moved.
    :type dir_s: str
    :param coord_format: The file extension or format of coordinate files to be processed.
    :type coord_format: tuple
    :param with_img: A flag indicating whether to include associated image files during
        dataset organization.
    :type with_img: bool
    :param img_format: The file extension or format of image files to be processed if
        `with_img` is True. Optional.
    :type img_format: tuple, optional
    :return: None
    :rtype: None

    :raises TardisError: If no coordinate files matching `coord_format` are found in the
        given directory.
    :raises TardisError: If `with_img` is True but no image files matching `img_format`
        are found in the given directory.
    """
    if coord_format == ".txt":
        area_list = [d for d in listdir(dir_s) if isdir(join(dir_s, d))]
        area_list = [f for f in area_list if not f.startswith((".", "train", "test"))]

        for i in area_list:
            copytree(join(dir_s, i), join(dir_s, "train", "masks", i))

    if not len([f for f in listdir(dir_s) if f.endswith(coord_format)]) > 0:
        TardisError(
            "121",
            "tardis_em/utils/dataset.py",
            f"No coordinate file found in given dir {dir_s}",
        )

    idx_coord = [f for f in listdir(dir_s) if f.endswith(coord_format)]

    for i in idx_coord:
        copyfile(src=join(dir_s, i), dst=join(dir_s, "train", "masks", i))

    """Sort coord with images if included"""
    if with_img:
        if not len([f for f in listdir(dir_s) if f.endswith(img_format)]) > 0:
            TardisError(
                "121",
                "tardis_em/utils/dataset.py",
                f"No image file found in given dir {dir_s}",
            )

        idx_coord = [f for f in listdir(dir_s) if f.endswith(img_format)]

        for i in idx_coord:
            copyfile(src=join(dir_s, i), dst=join(dir_s, "train", "imgs", i))


def build_test_dataset(dataset_dir: str, dataset_no: int, stanford=False):
    """
    Builds a test dataset by reorganizing and moving files from the train dataset directory
    to a test dataset directory based on specific selection logic. This function handles
    dataset creation for both the general case and a special case for Stanford datasets.
    Images and corresponding masks are moved into a new test directory, ensuring the train
    directory is properly split into train and test datasets.

    :param dataset_dir: Path to the directory containing the dataset.
    :type dataset_dir: str
    :param dataset_no: Number of datasets or partitions to consider when splitting for test data.
    :type dataset_no: int
    :param stanford: If True, applies specific folder organization and file movement logic
                     for Stanford datasets. Defaults to False.
    :type stanford: bool

    :return: None
    """
    if "test" not in listdir(dataset_dir) and "train" not in listdir(dataset_dir):
        TardisError(
            "122",
            "tardis_em/utils/dataset.py",
            f"Could not find train or test folder in directory {dataset_dir}",
        )

    if stanford:
        mkdir(join(dataset_dir, "test", "masks", "Area_1"))
        mkdir(join(dataset_dir, "test", "masks", "Area_3"))
        mkdir(join(dataset_dir, "test", "masks", "Area_5"))

        move(
            join(dataset_dir, "train", "masks", "Area_1", "office_1"),
            join(dataset_dir, "test", "masks", "Area_1", "office_1"),
        )
        move(
            join(dataset_dir, "train", "masks", "Area_3", "WC_1"),
            join(dataset_dir, "test", "masks", "Area_3", "WC_1"),
        )
        move(
            join(dataset_dir, "train", "masks", "Area_5", "storage_1"),
            join(dataset_dir, "test", "masks", "Area_5", "storage_1"),
        )

    image_list = listdir(join(dataset_dir, "train", "imgs"))
    image_list = [i for i in image_list if i.endswith((".tif", ".mrc"))]

    if len(image_list) < 1000:
        # Get list of dataset
        images = []
        for i in range(dataset_no):
            df_imgs = [img for img in image_list if img.startswith(f"{i}")]
            images.append(df_imgs)

        # For each image select 20% random patches
        images = [
            i
            for id_, i in enumerate(images)
            if id_ in random.sample(range(0, len(images)), int(len(images) // 5))
        ]

        for i in images:
            list_move = []
            for j in random.sample(range(0, len(i) - 1), 4 if len(i) > 10 else 0):
                list_move.append(i[j])

            for j in list_move:
                shutil.move(
                    join(dataset_dir, "train", "imgs", j),
                    join(dataset_dir, "test", "imgs", j),
                )

                shutil.move(
                    join(dataset_dir, "train", "masks", j[:-4] + "_mask" + j[-4:]),
                    join(dataset_dir, "test", "masks", j[:-4] + "_mask" + j[-4:]),
                )
    else:
        if len(image_list) > 62500:
            images = random.sample(range(0, len(image_list)), 1000)
        else:
            images = random.sample(
                range(0, len(image_list)), int(len(image_list) // 25)
            )

        for i in images:
            j = image_list[i]

            shutil.move(
                join(dataset_dir, "train", "imgs", j),
                join(dataset_dir, "test", "imgs", j),
            )
            shutil.move(
                join(dataset_dir, "train", "masks", j[:-4] + "_mask" + j[-4:]),
                join(dataset_dir, "test", "masks", j[:-4] + "_mask" + j[-4:]),
            )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from typing import Tuple, Union, List

import numpy as np
import torch

from tardis_em.dist_pytorch.datasets.augmentation import BuildGraph
from tardis_em.utils.errors import TardisError


class PatchDataSet:
    """
    Represents a class designed for configuring and processing point cloud data. This includes
    functionalities for down-sampling points, defining overlapping feature patches, and specifying
    output formats such as tensor representation or graph structure. It also initializes and manages
    patch size configurations.
    """

    def __init__(
        self,
        max_number_of_points=500,
        overlap=0.15,
        drop_rate=None,
        graph=True,
        tensor=True,
    ):
        """
        Class that initializes the configuration for point cloud down-sampling and patch processing. It sets up
        parameters for down-sampling the point cloud, configuring the feature patches, and determining output
        formats such as tensor or graph. It also provides initial space for the patch size in the configuration.

        :param max_number_of_points: The maximum number of points for point cloud down-sampling.
        :param overlap: A floating-point value specifying the overlap percentage
            for creating patches.
        :param drop_rate: Floating-point value representing the drop rate for feature reduction.
        :param graph: Boolean to enable or disable graph-based output processing.
        :param tensor: Boolean to enable or disable tensor-based output processing.
        """
        # Point cloud down-sampling setting
        self.DOWNSAMPLING_TH = max_number_of_points

        # Patch setting
        self.voxel = None
        self.drop_rate = drop_rate
        self.TORCH_OUTPUT = tensor
        self.GRAPH_OUTPUT = graph
        self.EXPAND = 0.1  # Expand the boundary box by 10%
        self.STRIDE = overlap  # Create 15% overlaps between patches

        # Initialization
        self.INIT_PATCH_SIZE = np.zeros((2, 3))

    def boundary_box(self, coord, offset=None) -> np.ndarray:
        """
        Calculate the boundary box for the given coordinates, optionally applying an
        offset. The function computes the minimum and maximum dimensions along each
        axis (x, y, z if applicable). If the offset is provided, the boundary box is
        expanded proportionally around the center. The result is a NumPy array
        representing the boundary box with two points: the lower and upper bounds.

        :param coord: The input array of coordinates where rows represent points, and
                      columns represent dimensional axes (x, y, and optionally z).
        :type coord: np.ndarray
        :param offset: An optional expansion multiplier for the bounding box dimensions.
                       If provided, it scales the boundary box's size relative to its
                       center.
        :type offset: float, optional
        :return: A NumPy array representing the boundary box with two rows: the minimum
                 (x, y, z) coordinates and the maximum (x, y, z) coordinates.
        :rtype: np.ndarray
        """
        # Define x,y and (z) min and m
        # ax sizes
        if coord.shape[1] == 3:
            min_x, min_y, min_z = np.min(coord, axis=0)
            max_x, max_y, max_z = np.max(coord, axis=0)
        else:
            min_x, min_y = np.min(coord, axis=0)
            max_x, max_y = np.max(coord, axis=0)
            min_z, max_z = 0, 0

        if offset is not None:
            self.EXPAND = offset

        dx = ((min_x + max_x) / 2) - min_x
        min_x, max_x = min_x - dx * self.EXPAND, max_x + dx * self.EXPAND

        dy = ((min_y + max_y) / 2) - min_y
        min_y, max_y = min_y - dy * self.EXPAND, max_y + dy * self.EXPAND

        dz = ((min_z + max_z) / 2) - min_z
        min_z, max_z = min_z - dz * self.EXPAND, max_z + dz * self.EXPAND

        dx, dy = abs(min_x - max_x), abs(min_y - max_y)
        d = dx if max(dx, dy) == 0 else dy

        self.drop_rate = d * 0.01
        return np.array([(min_x, min_y, min_z), (max_x, max_y, max_z)])

    @staticmethod
    def center_patch(bbox, voxel_size=1) -> np.ndarray:
        """
        Calculates the coordinates for the centers of voxels within a given bounding box.
        The bounding box is divided into a three-dimensional grid using the specified voxel size.
        The method ensures a minimum of 2 voxels along each axis to avoid degenerate cases.
        The voxel centers are returned as a NumPy array containing their coordinates.

        :param bbox: The bounding box defined as a 2x3 NumPy array where the first row contains
            the minimum x, y, z coordinates and the second row contains the maximum x, y, z
            coordinates.
        :type bbox: np.ndarray
        :param voxel_size: The size of each voxel along all dimensions. Defaults to 1.
        :type voxel_size: float
        :return: A NumPy array with the coordinates of all voxel centers as rows.
        :rtype: np.ndarray
        """
        # Calculate the number of voxels along each axis
        n_x = int(np.ceil((bbox[1, 0] - bbox[0, 0]) / voxel_size))
        n_y = int(np.ceil((bbox[1, 1] - bbox[0, 1]) / voxel_size))
        n_z = int(np.ceil((bbox[1, 2] - bbox[0, 2]) / voxel_size)) // 2

        # Calculate the coordinates of the voxel centers
        if n_x < 2:
            n_x = 2
        x = np.linspace(bbox[0, 0] - voxel_size / 2, bbox[1, 0] + voxel_size / 2, n_x)

        if n_y < 2:
            n_y = 2
        y = np.linspace(bbox[0, 1] - voxel_size / 2, bbox[1, 1] + voxel_size / 2, n_y)

        if n_z < 2:
            n_z = 2
        z = np.linspace(bbox[0, 2] - voxel_size / 2, bbox[1, 2] + voxel_size / 2, n_z)

        xv, yv, zv = np.meshgrid(x, y, z, indexing="ij")
        voxel_centers = np.column_stack((xv.flatten(), yv.flatten(), zv.flatten()))

        return voxel_centers

    def points_in_patch(self, coord: np.ndarray, patch_center: np.ndarray) -> bool:
        """
        Determines if specific points represented by their coordinates are located within a
        defined patch based on the patch's center and pre-defined size. The patch is modeled
        as a cuboid, and each coordinate is verified to be within the patch bounds along
        all axes.

        :param coord: A numpy array of shape (N, 3), where N is the number of points,
            and each row represents the x, y, z coordinates of a point.
        :param patch_center: A numpy array of shape (3,), representing the x, y, z
            coordinates of the patch center.
        :return: A boolean array of shape (N,), where each element signifies whether
            the corresponding point in `coord` is located within the defined patch limits.
        :rtype: np.ndarray
        """
        patch_size_x = self.INIT_PATCH_SIZE[0]
        patch_size_y = self.INIT_PATCH_SIZE[1]
        patch_size_z = self.INIT_PATCH_SIZE[2]

        # Bounding box for patch center
        patch_x, patch_y, patch_z = patch_center
        patch_min_x, patch_max_x = patch_x - patch_size_x, patch_x + patch_size_x
        patch_min_y, patch_max_y = patch_y - patch_size_y, patch_y + patch_size_y
        patch_min_z, patch_max_z = patch_z - patch_size_z, patch_z + patch_size_z

        coord_idx = (
            (coord[:, 0] <= patch_max_x)
            & (coord[:, 0] >= patch_min_x)
            & (coord[:, 1] <= patch_max_y)
            & (coord[:, 1] >= patch_min_y)
            & (coord[:, 2] <= patch_max_z)
            & (coord[:, 2] >= patch_min_z)
        )

        return coord_idx

    def optimal_patches(self, coord: np.ndarray, random=False) -> List[bool]:
        """
        The `optimal_patches` function calculates and returns a list of boolean
        values representing whether specific patches within a given coordinate
        space contain points that meet certain criteria. It operates in two
        modes: random patch selection or systematic grid-based patch calculation.

        This function first calculates a bounding box for the input coordinates.
        It determines voxel sizes dynamically based on the bounding box dimensions,
        applies strategies for patch creation, and evaluates patches for spatial
        distribution of points. The two modes allow either randomly selected patches
        or systematic patches using voxel downscaling until a threshold is satisfied.

        :param coord: The spatial coordinate input as a numpy array.
        :type coord: np.ndarray

        :param random: Determines if the patch selection should follow
            a random process or systematic method. Default is False.
        :type random: bool, optional

        :return: A list of boolean values indicating which patches meet
            the specific criteria.
        :rtype: List[bool]
        """
        bbox = self.boundary_box(coord)
        if self.voxel is not None:
            voxel = self.voxel + self.drop_rate
        else:
            voxel = abs(bbox[0, 0] - bbox[1, 0]) // 10 + self.drop_rate
            if voxel == 0:
                voxel = abs(bbox[0, 0] - bbox[1, 0]) + self.drop_rate

        all_patch = []

        """ Find points index in patches """
        if random:
            if self.voxel is not None:
                voxel = self.voxel + self.drop_rate
            else:
                voxel = abs(bbox[0, 0] - bbox[1, 0]) // 10
                if voxel == 0:
                    voxel = abs(bbox[0, 0] - bbox[1, 0]) + self.drop_rate

            patch_grid = self.center_patch(bbox=bbox, voxel_size=voxel)

            x_pos = np.sort(np.unique(patch_grid[:, 0]))
            if len(x_pos) > 1:
                x_pos = (x_pos[1] - x_pos[0]) / 2
            else:
                x_pos = bbox[[1, 0]] - x_pos[0]

            y_pos = np.sort(np.unique(patch_grid[:, 1]))
            if len(y_pos) > 1:
                y_pos = (y_pos[1] - y_pos[0]) / 2
            else:
                y_pos = bbox[1, 1] - y_pos[0]

            z_pos = np.sort(np.unique(patch_grid[:, 2]))
            if len(z_pos) > 1:
                z_pos = (z_pos[1] - z_pos[0]) / 2
            else:
                z_pos = bbox[1, 2] - z_pos[0]

            self.INIT_PATCH_SIZE = [
                x_pos + (x_pos * self.STRIDE),
                y_pos + (y_pos * self.STRIDE),
                z_pos + (z_pos * self.STRIDE),
            ]

            """Find points in each patch"""
            for patch in patch_grid:
                point_idx = self.points_in_patch(coord=coord, patch_center=patch)
                all_patch.append(point_idx)
            all_patch_df = [patch for patch in all_patch if np.sum(patch) > 0]
            all_patch_bool = [
                True if np.sum(patch) > 0 else False for patch in all_patch
            ]

            """Pick random patch"""
            df_all = []
            for _ in range(10):
                # Initially random pick
                random_ = np.argwhere(all_patch_bool).flatten()[
                    np.random.choice(len(all_patch_df))
                ]
                all_patch = [
                    self.points_in_patch(coord=coord, patch_center=patch_grid[random_])
                ]

                # Check if picked voxel have more than self.mesh points
                # and less than downsample threshold
                pc_size = np.sum(all_patch[0])

                while (
                    pc_size < self.DOWNSAMPLING_TH * 0.5
                    or pc_size > self.DOWNSAMPLING_TH
                ):
                    if pc_size > self.DOWNSAMPLING_TH:
                        while pc_size > self.DOWNSAMPLING_TH:
                            self.INIT_PATCH_SIZE = [
                                self.INIT_PATCH_SIZE[0] * 0.99,
                                self.INIT_PATCH_SIZE[1] * 0.99,
                                self.INIT_PATCH_SIZE[2] * 0.99,
                            ]
                            all_patch = [
                                self.points_in_patch(
                                    coord=coord, patch_center=patch_grid[random_]
                                )
                            ]
                            pc_size = np.sum(all_patch[0])
                    elif pc_size < self.DOWNSAMPLING_TH * 0.5:
                        while pc_size < self.DOWNSAMPLING_TH * 0.8:
                            self.INIT_PATCH_SIZE = [
                                self.INIT_PATCH_SIZE[0] * 1.01,
                                self.INIT_PATCH_SIZE[1] * 1.01,
                                self.INIT_PATCH_SIZE[2] * 1.01,
                            ]
                            all_patch = [
                                self.points_in_patch(
                                    coord=coord, patch_center=patch_grid[random_]
                                )
                            ]
                            pc_size = np.sum(all_patch[0])

                    if (
                        pc_size > self.DOWNSAMPLING_TH
                        or pc_size < self.DOWNSAMPLING_TH * 0.5
                    ):
                        random_ = np.argwhere(all_patch_bool).flatten()[
                            np.random.choice(len(all_patch_df))
                        ]
                        all_patch = [
                            self.points_in_patch(
                                coord=coord, patch_center=patch_grid[random_]
                            )
                        ]
                        pc_size = np.sum(all_patch[0])
                df_all.append(all_patch[0])
            all_patch = df_all
        else:
            th = 1
            while th != 0:
                all_patch = []

                """Initialize search with new voxel size"""
                voxel = round(voxel - self.drop_rate, 1)
                patch_grid = self.center_patch(bbox=bbox, voxel_size=voxel)
                if len(patch_grid) < 2:
                    continue

                x_pos = np.sort(np.unique(patch_grid[:, 0]))
                if len(x_pos) > 1:
                    x_pos = (x_pos[1] - x_pos[0]) / 2
                else:
                    x_pos = bbox[[1, 0]] - x_pos[0]

                y_pos = np.sort(np.unique(patch_grid[:, 1]))
                if len(y_pos) > 1:
                    y_pos = (y_pos[1] - y_pos[0]) / 2
                else:
                    y_pos = bbox[1, 1] - y_pos[0]

                z_pos = np.sort(np.unique(patch_grid[:, 2]))
                if len(z_pos) > 1:
                    z_pos = (z_pos[1] - z_pos[0]) / 2
                else:
                    z_pos = bbox[1, 2] - z_pos[0]

                self.INIT_PATCH_SIZE = [
                    x_pos + (x_pos * self.STRIDE),
                    y_pos + (y_pos * self.STRIDE),
                    z_pos + (z_pos * self.STRIDE),
                ]

                """Find points in each patch"""
                for patch in patch_grid:
                    point_idx = self.points_in_patch(coord=coord, patch_center=patch)
                    all_patch.append(point_idx)

                all_patch = [patch for patch in all_patch if np.sum(patch) > 0]
                th = sum([True for p in all_patch if np.sum(p) > self.DOWNSAMPLING_TH])

            """ Combine smaller patches with threshold limit """
            new_patch = []
            while len(all_patch) > 0:
                df = all_patch[0]

                if np.sum(df) >= self.DOWNSAMPLING_TH:
                    new_patch.append(df)
                    all_patch.pop(0)
                else:
                    while np.sum(df) <= self.DOWNSAMPLING_TH:
                        if len(all_patch) == 1:
                            break
                        if np.sum(df) + np.sum(all_patch[1]) > self.DOWNSAMPLING_TH:
                            break
                        df += all_patch[1]
                        all_patch.pop(1)
                    new_patch.append(df)
                    all_patch.pop(0)
            all_patch = new_patch

        return all_patch

    @staticmethod
    def normalize_idx(coord_with_idx: np.ndarray) -> np.ndarray:
        """
        Normalizes the index values of coordinates in the given array. The method updates the array
        such that indices in the first column are replaced with normalized indices, starting from zero.
        The normalization process maps unique indices to a consecutive range of indices.

        :param coord_with_idx: A numpy array where the first column contains coordinate indices to be
            normalized.
        :type coord_with_idx: np.ndarray
        :return: The input array with normalized indices in its first column.
        :rtype: np.ndarray
        """
        unique_idx, inverse_idx = np.unique(coord_with_idx[:, 0], return_inverse=True)
        # Hot-fix for numpy 2.0.0
        if inverse_idx.ndim == 2:
            inverse_index = inverse_idx[:, 0]

        norm_idx = np.arange(len(unique_idx))

        for _, id_ in enumerate(unique_idx):
            mask = coord_with_idx[:, 0] == id_
            coord_with_idx[:, 0][mask] = norm_idx[inverse_idx[mask]]
        return coord_with_idx

    def output_format(self, data: np.ndarray) -> Union[np.ndarray, torch.Tensor]:
        """
        Converts the given data into the desired output format. If the `TORCH_OUTPUT`
        flag is set, the input NumPy array is converted to a PyTorch tensor of type
        `torch.float32`.

        :param data: Input data as a NumPy array.
        :type data: np.ndarray
        :return: The input data converted to the desired format. Returns a PyTorch
            tensor if `TORCH_OUTPUT` is enabled, otherwise returns the original
            NumPy array.
        :rtype: Union[np.ndarray, torch.Tensor]
        """
        if self.TORCH_OUTPUT:
            data = torch.from_numpy(data).type(torch.float32)
        return data

    def patched_dataset(
        self,
        coord: np.ndarray,
        label_cls=None,
        rgb=None,
        mesh=6,
        random=False,
        voxel_size=None,
    ) -> Union[Tuple[List, List, List, List, List], Tuple[List, List, List, List]]:
        """
        Generates patches of spatial data and optional associated properties (class labels,
        graph structures, and RGB values) for deep learning tasks. This function processes
        input spatial coordinates to divide them into smaller, manageable patches, and
        conditionally builds graph structures or assigns additional information like class
        labels and RGB values for each patch. This method supports either a 2D-to-3D
        transformation or segmented and raw data processing, depending on the provided input
        configurations.

        :param coord: An array of spatial coordinates, potentially in 2D, 3D, or segmented
            4D form.
        :param label_cls: Optional array of class labels for spatial coordinates.
        :param rgb: Optional array of RGB values for spatial coordinates.
        :param mesh: An integer indicating the scale or structure of graph connectivity
            (e.g., K in a K-Nearest Neighbors graph).
        :param random: Boolean indicating whether to randomly segment patches or determine
            them based on an optimization algorithm.
        :param voxel_size: Optional numeric value specifying the desired voxel size
            for downsampling spatial data.

        :return: A tuple of lists with various outputs segmented into patches:
            - Point cloud patches for input coordinates.
            - RGB data patches if available.
            - Graph patches if graph output is configured.
            - Indices of original data points corresponding to each patch.
            - Optional class label patches if `label_cls` is provided.
        """
        coord_patch = []
        graph_patch = []
        output_idx = []

        if voxel_size is not None:
            self.voxel = voxel_size
        else:
            self.voxel = None

        if self.GRAPH_OUTPUT:
            if coord.shape[1] not in [3, 4]:
                TardisError(
                    "113",
                    "tardis_em/dist_pytorch/datasets/patches.py",
                    "If graph True, coord must by of shape"
                    f"[Dim x X x Y x (Z)], but is: {coord.shape}",
                )
            segmented_coord = coord
            coord = coord[:, 1:]

            if mesh > 2:
                graph_builder = BuildGraph(K=mesh, mesh=True)
            else:
                graph_builder = BuildGraph(K=mesh, mesh=False)
        else:
            graph_builder = None
            if coord.shape[1] not in [2, 3]:
                TardisError(
                    "113",
                    "tardis_em/dist_pytorch/datasets/patches.py",
                    "If graph False, coord must by of shape"
                    f"[X x Y x (Z)], but is: {coord.shape}",
                )
            segmented_coord = None
            if coord.shape[1] == 2:
                """Transform 2D coord to 3D of shape [X, Y, Z]"""
                coord = np.vstack(
                    (
                        coord[:, 0],
                        coord[:, 1],
                        np.zeros((coord.shape[0],)),
                    )
                ).T
            else:
                coord = coord

        if coord.shape[0] <= self.DOWNSAMPLING_TH:
            coord_ds = [True for _ in list(range(0, coord.shape[0], 1))]

            """ Build point cloud for each patch """
            coord_patch.append(self.output_format(coord[coord_ds, :]))

            """ Optionally - Build graph for each patch """
            if self.GRAPH_OUTPUT:
                coord_label = segmented_coord[coord_ds, :]
                coord_label = self.normalize_idx(coord_label)

                graph_patch.append(self.output_format(graph_builder(coord=coord_label)))

            """ Build output index for each patch """
            output_idx.append(np.where(coord_ds)[0])

            """ Build class label index for each patch """
            if label_cls is not None:
                cls_patch = [label_cls]
            else:
                cls_patch = [self.output_format(np.zeros((1, 1)))]

            """ Build rgb node label index for each patch """
            if rgb is not None:
                rgb_patch = [rgb]
            else:
                rgb_patch = [self.output_format(np.zeros((1, 1)))]
        else:  # Build patches for PC with max num. of point per patch
            cls_patch = []
            rgb_patch = []

            """ Find points index in patches """
            all_patch = self.optimal_patches(coord=coord, random=random)

            """Build embedded feature per patch"""
            for i in all_patch:
                """Find points and optional images for each patch"""
                df_patch_keep = i

                df_patch = coord[df_patch_keep, :]
                output_df = np.where(df_patch_keep)[0]
                coord_ds = [True for _ in list(range(0, df_patch.shape[0], 1))]

                """ Build point cloud for each patch """
                coord_patch.append(self.output_format(df_patch[coord_ds, :]))

                """ Optionally - Build graph for each patch """
                if self.GRAPH_OUTPUT:
                    segment_patch = segmented_coord[df_patch_keep, :]
                    segment_patch = self.normalize_idx(segment_patch[coord_ds, :])

                    graph_patch.append(
                        self.output_format(graph_builder(coord=segment_patch))
                    )

                """ Build output index for each patch """
                output_idx.append(output_df[coord_ds])

                """ Build class label index for each patch """
                if label_cls is not None:
                    cls_df = label_cls[df_patch_keep]
                    cls_new = np.zeros((cls_df.shape[0], 200))
                else:
                    cls_df = [0]
                    cls_new = np.zeros((1, 200))

                for id_, j in enumerate(cls_df):
                    df = np.zeros((1, 200))
                    df[0, int(j)] = 1
                    cls_new[id_, :] = df
                cls_patch.append(self.output_format(cls_new))

                """ Build rbg node label index for each patch """
                if rgb is not None:
                    rgb_df = rgb[df_patch_keep]
                else:
                    rgb_df = np.zeros((1, 1))

                rgb_patch.append(self.output_format(rgb_df))

        if self.GRAPH_OUTPUT:
            return coord_patch, rgb_patch, graph_patch, output_idx, cls_patch
        else:
            return coord_patch, rgb_patch, output_idx, cls_patch



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from typing import Iterable

import numpy as np

from tardis_em.utils.errors import TardisError


def interpolate_generator(points: np.ndarray) -> Iterable:
    """
    Generates interpolated points between two given 2D or 3D points. The function
    determines the interpolation path in a pixel-grid-like manner between the start
    and end points, based on Bresenham-like algorithms. Interpolation supports
    only 2D and 3D coordinate systems, and exactly two points must be provided for
    interpolation.

    :param points: An array of shape (2, 2) for 2D points or (2, 3) for 3D points.
                   Defines the two points between which interpolation is performed.
    :return: An iterable generator that yields tuples representing (x, y) in 2D
             space and (x, y, z) in 3D space.
    """
    if points.shape not in [(2, 3), (2, 2)]:
        TardisError(
            "134",
            "tardis_em/cnn/data_processing.md/interpolation.py",
            "Interpolation supports only 2D/3D for 2 points at a time; "
            f"But {points.shape} was given!",
        )

    points = np.round(points).astype(np.int32)
    if points.shape == (2, 2):
        dim_ = 2
    else:
        dim_ = 3

    # Collect first and last point in array for XYZ
    x0, x1 = points[0, 0], points[1, 0]
    y0, y1 = points[0, 1], points[1, 1]
    if dim_ == 2:
        z0, z1 = 0, 0
    else:
        z0, z1 = points[0, 2], points[1, 2]

    # Delta between first and last point to interpolate
    delta_x, delta_y, delta_z = x1 - x0, y1 - y0, z1 - z0

    # Calculate axis to iterate throw
    max_delta = np.where(
        (abs(delta_x), abs(delta_y), abs(delta_z))
        == np.max((abs(delta_x), abs(delta_y), abs(delta_z)))
    )[0][0]
    if delta_x == 0 and delta_y == 0 and delta_z == 0:
        max_delta = 3

    # Calculate scaling direction + or - or None
    dx_sign, dy_sign, dz_sign = np.sign(delta_x), np.sign(delta_y), np.sign(delta_z)

    # Calculating scaling threshold
    delta_err_x = (
        0.0
        if delta_x == 0
        else (
            abs(delta_x / delta_y)
            if delta_y != 0
            else abs(delta_x / delta_z) if delta_z != 0 else 0.0
        )
    )
    delta_err_y = (
        0.0
        if delta_y == 0
        else (
            abs(delta_y / delta_x)
            if delta_x != 0
            else abs(delta_y / delta_z) if delta_z != 0 else 0.0
        )
    )

    if dim_ != 2:
        delta_err_z = (
            0.0
            if delta_z == 0
            else (
                np.minimum(abs(delta_z / delta_x), abs(delta_z / delta_y))
                if delta_x != 0 and delta_y != 0
                else (
                    abs(delta_z / delta_y)
                    if delta_x == 0 and delta_y != 0
                    else abs(delta_z / delta_x) if delta_x != 0 else 0.0
                )
            )
        )

    # Zero out threshold
    error_x, error_y, error_z = 0, 0, 0
    x, y, z = x0, y0, z0

    if max_delta == 0:  # Scale XYZ by iterating throw X axis
        for x in range(x0, x1, dx_sign):
            if dim_ != 2:
                yield x, y, z
            else:
                yield x, y

            # Iteratively add and scale Y axis
            error_y = error_y + delta_err_y
            while error_y >= 0.5:
                y += dy_sign
                error_y -= 1

            if dim_ != 2:
                # Iteratively add and scale Z axis
                error_z = error_z + delta_err_z
                while error_z >= 0.5:
                    z += dz_sign
                    error_z -= 1
    if max_delta == 1:  # Scale XYZ by iterating throw Y axis
        for y in range(y0, y1, dy_sign):
            if dim_ != 2:
                yield x, y, z
            else:
                yield x, y

            # Iteratively add and scale X axis
            error_x = error_x + delta_err_x
            while error_x >= 0.5:
                x += dx_sign
                error_x -= 1

            if dim_ != 2:
                # Iteratively add and scale Z axis
                error_z = error_z + delta_err_z
                while error_z >= 0.5:
                    z += dz_sign
                    error_z -= 1
    if max_delta == 2:  # Scale XYZ by iterating throw Z axis
        for z in range(z0, z1, dz_sign):
            if dim_ != 2:
                yield x, y, z
            else:
                yield x, y
    if max_delta == 3:  # Nothing to do
        if dim_ != 2:
            yield x, y, z
        else:
            yield x, y


def interpolation(points: np.ndarray) -> np.ndarray:
    """
    Performs 3D interpolation for XYZ coordinates using a given interpolation
    generator over sequential point pairs, and appends the last point to the
    output in integer format. The function creates a continuous series of
    interpolated points connecting all input coordinates.

    :param points: An array of shape (N, 3) representing a sequence of 3D points.

    :return: An array of shape (M, 3) containing the interpolated 3D coordinates,
             where M >= N due to newly generated intermediate points.
    """
    new_coord = []
    for i in range(0, len(points) - 1):
        """3D interpolation for XYZ dimension"""
        new_coord.append(list(interpolate_generator(points[i : i + 2, :])))

    # Append last point
    new_coord.append(list(np.round(points[-1, :]).astype(np.int32)))

    return np.vstack(new_coord)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from typing import Optional

import torch
import torch.nn as nn

from tardis_em.cnn.model.init_weights import init_weights
from tardis_em.cnn.utils.utils import number_of_features_per_level


class EncoderBlock(nn.Module):
    """
    Represents an encoder block with convolutional modules, optional max-pooling, dropout, and
    attention features for deep learning architectures.

    This class constructs a configurable encoder block. The encoder block supports max-pooling
    layers, dropout layers, attention features, and employs convolutional modules with specific
    kernel sizes and padding. The components used in the block can be customized through
    component identifiers.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        conv_module,
        conv_kernel=3,
        max_pool=True,
        dropout: Optional[float] = None,
        pool_kernel=2,
        padding=1,
        components="3gcr",
        num_group=8,
        attn_features=False,
    ):
        """
        Initializes an encoder block with optional max pooling, dropout, and convolutional
        modules. The encoder block can also support attention features when enabled. The
        constructor sets up the required modules, components, and parameters based on the
        provided input arguments.

        :param in_ch: Number of input channels for this encoder block.
        :type in_ch: int
        :param out_ch: Number of output channels for this encoder block.
        :type out_ch: int
        :param conv_module: A convolutional module or block used within the encoder.
        :param conv_kernel: Size of the convolutional kernel (default is 3).
        :type conv_kernel: int, optional
        :param max_pool: If True, adds a max-pooling layer to the block (default is True).
        :type max_pool: bool, optional
        :param dropout: Dropout rate to be applied. If None, no dropout layer is added.
        :type dropout: float, optional
        :param pool_kernel: Size of the kernel for max-pooling operations (default is 2).
        :type pool_kernel: int, optional
        :param padding: Amount of padding to apply in convolutional layers (default is 1).
        :type padding: int, optional
        :param components: String identifier defining specific layer components to use
            (e.g., "3gcr").
        :type components: str
        :param num_group: Number of groups for grouped convolutions (default is 8).
        :type num_group: int, optional
        :param attn_features: Enables the attention feature if set to True (default is False).
        :type attn_features: bool, optional

        :raises TypeError: If `conv_module` is not callable.
        :raises ValueError: If incorrect component identifiers are provided in `components` or
            if arguments are incompatible with the selected configuration.
        """
        super(EncoderBlock, self).__init__()

        self.attn_features = attn_features
        self.dropout = dropout

        """Optionally, add maxpool"""
        if max_pool:
            if "3" in components:
                self.maxpool = nn.MaxPool3d(kernel_size=pool_kernel)
            elif "2" in components:
                self.maxpool = nn.MaxPool2d(kernel_size=pool_kernel)
        else:
            self.maxpool = None

        """Optionally, add dropout layer"""
        if dropout is not None:
            self.dropout_layer = nn.Dropout(p=dropout)

        """Build CNN block"""
        self.conv_module = conv_module(
            in_ch=in_ch,
            out_ch=out_ch,
            block_type="encoder",
            kernel=conv_kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        if self.attn_features:
            self.attn_conv = conv_module(
                in_ch=in_ch + out_ch,
                out_ch=out_ch,
                block_type="encoder",
                kernel=conv_kernel,
                padding=padding,
                components=components,
                num_group=1 if (in_ch + out_ch) % num_group != 0 else num_group,
            )

        """Initialise the blocks"""
        for m in self.children():
            init_weights(m)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Processes input tensor through layers including convolutional, attention mechanisms,
        optional pooling, and dropout, returning the transformed tensor. This method serves as
        the key forward computation logic for the module.

        :param x: Input tensor to be processed.
        :type x: torch.Tensor

        :return: Processed output tensor.
        :rtype: torch.Tensor
        """
        if self.maxpool is not None:
            x = self.maxpool(x)

        x_attn = self.conv_module(x)

        if self.attn_features:
            x_attn = self.attn_conv(torch.cat((x, x_attn), dim=1))

        if self.dropout is not None:
            x_attn = self.dropout_layer(x_attn)

        return x_attn


def build_encoder(
    in_ch: int,
    conv_layers: int,
    conv_layer_scaler: int,
    conv_kernel: int or tuple,
    padding: int or tuple,
    num_group: int,
    components: str,
    pool_kernel: int or tuple,
    conv_module,
    dropout: Optional[float] = None,
    attn_features=False,
) -> nn.ModuleList:
    """
    Constructs and returns a sequence of encoder blocks as a module list. Each encoder block
    is defined based on the provided parameters and forms a hierarchical structure of
    feature extraction layers. The first encoder block does not use max pooling, while
    subsequent layers include max pooling operations as specified. The method also
    accommodates advanced features such as attention mechanisms and grouped convolutions.

    :param in_ch: Number of input image channels for the first encoder block. Each subsequent
        block adapts based on feature scaling.
    :param conv_layers: Number of convolutional layers to create in the encoder architecture.
    :param conv_layer_scaler: Multiplier to compute the number of feature maps at each level
        of the encoder.
    :param conv_kernel: Kernel size for convolution operations in the encoder blocks. Can be
        an integer or tuple.
    :param padding: Padding value for the convolutional layers. Accepts integer or tuple for
        specific layer configurations.
    :param num_group: Number of groups for grouped convolution operations.
    :param components: String identifier or configuration for additional encoder components
        such as normalization layers or activation functions.
    :param pool_kernel: Kernel size used for max pooling operations across encoder blocks
        except the first block. Accepts integer or tuple.
    :param conv_module: A callable or module defining the specific implementation for
        convolutional operations.
    :param dropout: Optional dropout probability to apply regularization in each encoder
        block. Defaults to None, disabling dropout.
    :param attn_features: Boolean flag to enable or disable attention mechanisms in the encoder
        blocks. Defaults to False.

    :return: Returns a `nn.ModuleList` containing the constructed encoder blocks.
    """
    encoders = []
    feature_map = number_of_features_per_level(
        channel_scaler=conv_layer_scaler, num_levels=conv_layers
    )

    for i, feature in enumerate(feature_map):
        if i == 0:  # first encoder layer skips max pooling
            encoder = EncoderBlock(
                in_ch=in_ch,
                out_ch=feature,
                conv_module=conv_module,
                conv_kernel=conv_kernel,
                dropout=dropout,
                max_pool=False,
                padding=padding,
                components=components,
                num_group=num_group,
                attn_features=attn_features,
            )
        else:
            encoder = EncoderBlock(
                in_ch=feature_map[i - 1],
                out_ch=feature,
                conv_module=conv_module,
                conv_kernel=conv_kernel,
                dropout=dropout,
                pool_kernel=pool_kernel,
                padding=padding,
                components=components,
                num_group=num_group,
                attn_features=attn_features,
            )
        encoders.append(encoder)

    return nn.ModuleList(encoders)



# #####################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
# #####################################################################

from math import sqrt

import numpy as np

from tardis_em.utils.load_data import ImportDataFromAmira
from tardis_em_analysis.spatial_graph_utils import FilterSpatialGraph
from tardis_em_analysis.geometry_metrics import total_length, tortuosity
from tardis_em_analysis.filament_utils import sort_segment, reorder_segments_id


def test_reorder_segments():
    coord = ImportDataFromAmira("./tests/test_data/data_type/am3D.CorrelationLines.am")
    segments = coord.get_segmented_points() / coord.get_pixel_size()
    segments = segments.astype(np.int32)
    segments[np.where(segments[:, 0] == 2)[0], 0] = 5

    reorder_segment = reorder_segments_id(coord=segments, order_range=None)
    assert np.all(np.unique(reorder_segment[:, 0]) == [0, 1, 2])

    reorder_segment = reorder_segments_id(coord=segments, order_range=(3, 6))
    assert np.all(np.unique(reorder_segment[:, 0]) == [3, 4, 5])


def test_sort_segment():
    # Test case 1: Check if the function sorts the points in line order
    coord = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15]])
    expected_output = np.array(
        [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15]]
    )

    assert np.array_equal(sort_segment(coord), expected_output)

    # Test case 2: Check if the function returns an array of the same shape as the input
    coord = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [13, 14, 15]])
    assert sort_segment(coord).shape == coord.shape

    # Test case 3: Check if the function can handle an empty input
    coord = np.array([])
    assert sort_segment(coord).shape == (0,)


def test_total_length():
    coord = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])
    expected_length = 3 * sqrt(27)
    assert (
        total_length(coord) == expected_length
    ), f"For coord={coord}, expected length={expected_length} but got {total_length(coord)}"

    coord = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_length = 2 * sqrt(3)
    assert (
        total_length(coord) == expected_length
    ), f"For coord={coord}, expected length={expected_length} but got {total_length(coord)}"

    coord = np.array([[1, 1, 1]])
    expected_length = 0
    assert (
        total_length(coord) == expected_length
    ), f"For coord={coord}, expected length={expected_length} but got {total_length(coord)}"


def test_tortuosity():
    coord = np.array([[1, 2, 1], [2, 5, 3], [4, 6, 6], [8, 9, 6]])
    expected_tortuosity = round(sqrt(1.268), 3)
    result = round(tortuosity(coord), 3)
    assert (
        expected_tortuosity == result
    ), f"For coord={coord}, expected tortuosity={expected_tortuosity} but got {result}"

    coord = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    expected_tortuosity = 1
    assert (
        tortuosity(coord) == expected_tortuosity
    ), f"For coord={coord}, expected tortuosity={expected_tortuosity} but got {tortuosity(coord)}"

    coord = np.array([[1, 1, 1]])
    expected_tortuosity = 1
    assert (
        tortuosity(coord) == expected_tortuosity
    ), f"For coord={coord}, expected tortuosity={expected_tortuosity} but got {tortuosity(coord)}"


def test_FilterWrapper():
    filter = FilterSpatialGraph(connect_seg_if_closer_then=0, filter_short_segments=0)
    coord = np.array(
        [
            [0, 1, 2, 1],
            [0, 2, 5, 3],
            [0, 4, 6, 6],
            [0, 8, 9, 6],
            [1, 12, 10, 8],
            [1, 15, 12, 9],
            [1, 18, 16, 12],
            [1, 22, 25, 15],
            [10, 102, 100, 80],
            [10, 150, 102, 90],
            [10, 180, 106, 102],
            [10, 220, 205, 105],
        ]
    )
    expect = np.array(
        [
            [0, 1, 2, 1],
            [0, 2, 5, 3],
            [1, 12, 10, 8],
            [1, 15, 12, 9],
            [1, 18, 16, 12],
            [1, 22, 25, 15],
            [2, 102, 100, 80],
            [2, 150, 102, 90],
            [2, 180, 106, 102],
        ]
    )
    filtered_coord = filter(coord)
    assert np.all(filtered_coord.flatten() == expect.flatten())



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import glob
from os import listdir, mkdir, rename, chmod
from os.path import isdir, join
from shutil import rmtree
from typing import Union

from tardis_em.utils.errors import TardisError


def build_new_dir(dir_s: str):
    """
    Builds a directory for temporary output files and ensures required files
    are present within the specified directory. If the required conditions
    are not met or the output directory already exists, appropriate actions
    are taken.

    :param dir_s: Path to the base directory where the temporary directory will
                 be created.
    :type dir_s: str
    :return: None
    """
    """ Build temp files directory """
    output = join(dir_s, "output")
    image_list = glob.glob(dir_s + "/*.tif")
    if not len(image_list) > 0:
        TardisError(
            "12",
            "tardis_em/utils/setup_envir.py",
            "At least one .tif image has to be in the directory!",
        )

    if not isdir(output):
        mkdir(output)
    else:
        TardisError(
            "12",
            "tardis_em/utils/setup_envir.py",
            "Output directory already exist! Files moved to new output_old "
            "directory.",
        )

        rename(output, join(dir_s, "output_old"))
        mkdir(output)


def build_temp_dir(dir_s: str):
    """
    Creates necessary directories in the specified base directory to ensure the required
    structure for processing is in place. The function first checks for the existence of
    a "temp" subdirectory and its nested structure within the provided `dir_`. If the
    "temp" and respective subdirectories already exist, they will be reset to a clean
    state. If not, proper directories will be created as needed. Additionally, verifies
    and creates a top-level "Predictions" directory if it doesn't exist.

    :param dir_s: The base directory where the directories need to be created or reset.
    :type dir_s: str
    :return: None
    """
    if isdir(join(dir_s, "temp")):
        if isdir(join(dir_s, "temp", "Patches")) or isdir(
            join(dir_s, "temp", "Predictions")
        ):
            clean_up(dir_s=dir_s)

            mkdir(join(dir_s, "temp"))
            mkdir(join(dir_s, "temp", "Patches"))
            mkdir(join(dir_s, "temp", "Predictions"))
        else:
            mkdir(join(dir_s, "temp", "Patches"))
            mkdir(join(dir_s, "temp", "Predictions"))
    else:
        mkdir(join(dir_s, "temp"))
        mkdir(join(dir_s, "temp", "Patches"))
        mkdir(join(dir_s, "temp", "Predictions"))

    if not isdir(join(dir_s, "Predictions")):
        mkdir(join(dir_s, "Predictions"))
    chmod(join(dir_s, "Predictions"), 0o777)
    chmod(join(dir_s, "temp"), 0o777)
    chmod(join(dir_s, "temp", "Patches"), 0o777)


def clean_up(dir_s: str):
    """
    Removes a directory named "temp" and its subdirectories, including
    "Patches" and "Predictions", from the specified directory path. Ensures
    recursive deletion of all files and subdirectories inside the "temp"
    directory.

    :param dir_s: The path to the parent directory containing the "temp"
        directory to be cleaned up. The directory path must be provided as a
        string.
    :return: None. The function performs cleanup operations without
        returning a value.
    """
    if isdir(join(dir_s, "temp")):
        if isdir(join(dir_s, "temp", "Patches")):
            rmtree(join(dir_s, "temp", "Patches"))

        if isdir(join(dir_s, "temp", "Predictions")):
            rmtree(join(dir_s, "temp", "Predictions"))

        rmtree(join(dir_s, "temp"))


def check_dir(
    dir_s: str,
    train_img: str,
    train_mask: str,
    test_img: str,
    test_mask: str,
    with_img: bool,
    img_format: Union[tuple, str],
    mask_format: Union[tuple, str, None],
) -> bool:
    """
    Validates the structure and contents of a dataset directory to ensure proper
    organization for training and testing processes. This function checks for
    the existence of specific subdirectories (`train` and `test`), as well as for
    the presence and consistency of files (both images and masks) in the dataset.
    It ensures that the number of image files matches the number of mask files,
    and optionally validates file formats.

    :param dir_s: Root directory of the dataset to validate.
    :type dir_s: str
    :param train_img: Path to the subdirectory containing training images.
    :type train_img: str
    :param train_mask: Path to the subdirectory containing training masks.
    :type train_mask: str
    :param test_img: Path to the subdirectory containing test images.
    :type test_img: str
    :param test_mask: Path to the subdirectory containing test masks.
    :type test_mask: str
    :param with_img: Indicates whether the validation should check for the
        presence of both images and masks, or just masks alone.
    :type with_img: bool
    :param img_format: File extension(s) for image files. It can be either
        a tuple of extensions or a single extension as a string.
    :type img_format: Union[tuple, str]
    :param mask_format: File extension(s) for mask files. It can be a tuple
        of extensions, a single extension as a string, or None to disable
        mask format validation.
    :type mask_format: Union[tuple, str, None]
    :return: A boolean indicating whether the directory structure and
        contents meet the validation criteria.
    :rtype: bool
    """
    if mask_format is None:
        return True

    dataset_test = False
    if isdir(join(dir_s, "train")) and isdir(join(dir_s, "test")):
        dataset_test = True

        if with_img:
            # Check if train img and coord exist and have same files
            if isdir(train_img) and isdir(train_mask):
                if len(
                    [f for f in listdir(train_img) if f.endswith(img_format)]
                ) == len([f for f in listdir(train_mask) if f.endswith(mask_format)]):
                    if (
                        len([f for f in listdir(train_img) if f.endswith(img_format)])
                        == 0
                    ):
                        dataset_test = False
                else:
                    dataset_test = False

            # Check if test img and mask exist and have same files
            if isdir(test_img) and isdir(test_mask):
                if len([f for f in listdir(test_img) if f.endswith(img_format)]) == len(
                    [f for f in listdir(test_mask) if f.endswith(mask_format)]
                ):
                    if (
                        len([f for f in listdir(test_img) if f.endswith(img_format)])
                        == 0
                    ):
                        dataset_test = False
                else:
                    dataset_test = False
        else:
            if isdir(train_img) and isdir(train_mask):
                if len([f for f in listdir(train_mask) if f.endswith(mask_format)]) > 0:
                    pass
                else:
                    dataset_test = False
            else:
                dataset_test = False

            if isdir(test_img) and isdir(test_mask):
                if len([f for f in listdir(test_mask) if f.endswith(mask_format)]) > 0:
                    pass
                else:
                    dataset_test = False
            else:
                dataset_test = False
    return dataset_test



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import time
from os import makedirs
from os.path import isdir, join
from shutil import rmtree
from typing import List

import numpy as np
import torch

from tardis_em.dist_pytorch.datasets.dataloader import build_dataset
from tardis_em.dist_pytorch.utils.segment_point_cloud import PropGreedyGraphCut
from tardis_em.cnn.datasets.build_dataset import (
    build_train_dataset,
)
from tardis_em.cnn.datasets.dataloader import PredictionDataset
from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import load_image
from tardis_em.utils.logo import print_progress_bar, TardisLogo
from tardis_em.utils.metrics import AP, AUC, calculate_f1, IoU, mcov
from tardis_em.utils.predictor import Predictor


class CnnBenchmark:
    """
    Wrapper for CNN benchmark dataset

    Args:
        model (Predictor): Predictor class with evaluated model.
        dataset (str): Dataset name to evaluate model on.
        dir_s (str): Dataset directory.
        patch_size (int): Image patch size for CNN model.
        threshold (float): Threshold value used for prediction..
    """

    def __init__(
        self,
        model: Predictor,
        dataset: str,
        dir_s: str,
        patch_size: int,
        threshold: float,
    ):
        self.tardis_progress = TardisLogo()
        self.title = "TARDIS - CNN Benchmark"
        self.tardis_progress(title=self.title)

        self.model = model
        self.threshold = threshold
        self.metric = {
            "Acc": [],
            "Prec": [],
            "Recall": [],
            "F1": [],
            "AUC": [],
            "IoU": [],
            "AP": [],
        }

        self.data_set = dataset
        self.dir = dir_s
        if isdir(join(dir_s, "train")):
            rmtree(join(dir_s, "train"))

        makedirs(join(dir_s, "train", "imgs"))
        makedirs(join(dir_s, "train", "masks"))

        if self.data_set in ["MT", "Mem"]:
            build_train_dataset(
                dataset_dir=self.dir,
                circle_size=150,
                resize_pixel_size=25,
                trim_xy=patch_size,
                trim_z=patch_size,
            )
        else:
            TardisError(
                id_="",
                py="tardis_em/benchmarks/benchmarks.py",
                desc=f"Given data set {self.data_set} is not supporter!",
            )

        self.eval_data = PredictionDataset(img_dir=join(self.dir, "train", "imgs"))

    def _benchmark(self, logits: np.ndarray, target: np.ndarray):
        input = np.where(logits >= self.threshold, 1, 0).astype(np.uint8)

        # ACC, Prec, Recall, F1
        accuracy_score, precision_score, recall_score, F1_score = calculate_f1(
            input, target, False
        )
        self.metric["Acc"].append(accuracy_score)
        self.metric["Prec"].append(precision_score)
        self.metric["Recall"].append(recall_score)
        self.metric["F1"].append(F1_score)

        # AUC
        self.metric["AUC"].append(AUC(logits, target))

        # IoU
        self.metric["IoU"].append(IoU(input, target))

        # AP
        self.metric["AP"].append(AP(logits, target))

    def _predict(self, input: torch.Tensor) -> np.ndarray:
        return self.model.predict(input[None, :])

    def _output_metric(self):
        return {k: np.mean(v) for k, v in self.metric.items()}

    def __call__(self) -> dict:
        """
        Main call for benchmarking CNN.

        It takes model, predict given image and run benchmark on it to return
        standard metrics for CNN.
        """
        iter_ = 1
        self.tardis_progress(
            title=self.title,
            text_1=f"Running image segmentation benchmark on "
            f"{self.data_set} dataset",
            text_4="Benchmark: In progress...",
            text_5="Current Task: CNN Benchmark...",
            text_8=print_progress_bar(0, len(self.eval_data)),
        )

        for i in range(len(self.eval_data)):
            """Predict"""
            # Pick image['s]
            input, name = self.eval_data.__getitem__(i)

            if i == 0:
                start = time.time()
                input = self._predict(input)
                end = time.time()

                # Scale progress bar refresh to 10s
                iter_ = 10 // (end - start)
                if iter_ < 1:
                    iter_ = 1
            else:
                input = self._predict(input)

            target, _ = load_image(join(self.dir, "train", "masks", f"{name}_mask.tif"))
            target = target.astype(np.uint8)

            """Benchmark"""
            self._benchmark(input, target)

            metric_1 = (
                f'mAcc: {round(np.mean(self.metric["Acc"]), 2)}; '
                f'mPrec: {round(np.mean(self.metric["Prec"]), 2)}; '
                f'mRecall: {round(np.mean(self.metric["Recall"]), 2)}; '
                f'mF1: {round(np.mean(self.metric["F1"]), 2)}'
            )
            metric_2 = (
                f'mAUC: {round(np.mean(self.metric["AUC"]), 2)}; '
                f'mIoU: {round(np.mean(self.metric["IoU"]), 2)}; '
                f'mAP: {round(np.mean(self.metric["AP"]), 2)}'
            )

            if i % iter_ == 0:
                # Tardis progress bar update
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Running image segmentation benchmark on "
                    f"{self.data_set} dataset",
                    text_4="Benchmark: In progress...",
                    text_5="Current Task: CNN Benchmark...",
                    text_6=metric_1,
                    text_7=metric_2,
                    text_8=print_progress_bar(i, len(self.eval_data)),
                )

        rmtree(join(self.dir, "train"))
        return self._output_metric()


class DISTBenchmark:
    """
    Wrapper for DIST benchmark dataset

    Args:
        model (Predictor): Predictor class with evaluated model.
        dataset (str): Dataset name to evaluate model on.
        dir_s (str): Dataset directory.
        points_in_patch (int): Max. number of points in sub-graphs.
        threshold (float): Threshold value used for prediction..
    """

    def __init__(
        self,
        model: Predictor,
        dataset: str,
        dir_s: str,
        points_in_patch: int,
        threshold: float,
    ):
        self.tardis_progress = TardisLogo()
        self.title = "TARDIS - DIST Benchmark"
        self.tardis_progress(title=self.title)

        self.model = model
        self.dir = dir_s
        self.data_set = dataset
        if self.data_set in ["MT", "Mem"]:
            self.sort = True
        else:
            self.sort = False

        self.threshold = threshold
        self.metric = {
            "IoU": [],  # Graph
            "AUC": [],  # Graph
            "mCov": [],  # Instance
            "mWCov": [],  # Instance
        }
        if dataset.endswith("rgb"):
            self.rgb = True
        else:
            self.rgb = False

        self.eval_data = build_dataset(
            dataset_type=dataset,
            dirs=[None, self.dir],
            max_points_per_patch=points_in_patch,
            benchmark=True,
        )

        if dataset in ["MT", "Mem"]:
            self.max_connections = 2
        else:
            self.max_connections = 4

    def _benchmark_graph(self, logits: np.ndarray, target: np.ndarray):
        input = np.where(logits >= self.threshold, 1, 0).astype(np.uint8)

        # IoU
        self.metric["IoU"].append(IoU(input, target, True))

        # AUC
        self.metric["AUC"].append(AUC(logits, target, True))

    def _benchmark_IS(
        self, logits: List[np.ndarray], coords: np.ndarray, output_idx: List[np.ndarray]
    ):
        # Graph cut
        GraphToSegment = PropGreedyGraphCut(
            threshold=self.threshold, connection=self.max_connections
        )
        input_IS = GraphToSegment.patch_to_segment(
            graph=logits, coord=coords[:, 1:], idx=output_idx, prune=5, sort=self.sort
        )

        # mCov and mWCov
        mCov, mwCov = mcov(input_IS, coords)
        self.metric["mCov"].append(mCov)
        self.metric["mWCov"].append(mwCov)

    def _predict(self, input, node=None):
        if node is not None:
            return self.model.predict(input, node)
        return self.model.predict(input)

    def _output_metric(self):
        return {k: np.mean(v) for k, v in self.metric.items()}

    def _update_metric_pg(self):
        mean_or_nan = lambda x: round(np.mean(x), 2) if len(x) != 0 else "nan"

        iou = mean_or_nan(self.metric["IoU"])
        auc = mean_or_nan(self.metric["AUC"])
        m_mcov = mean_or_nan(self.metric["mCov"])
        m_mwcov = mean_or_nan(self.metric["mWCov"])

        pg = f"IoU: {iou}; " f"AUC: {auc}; " f"mCov: {m_mcov}; " f"mWCov: {m_mwcov}"
        return pg

    def __call__(self):
        # Tardis progress bar update
        self.tardis_progress(
            title=self.title,
            text_1=f"Running point cloud segmentation benchmark on " f"{self.data_set}",
            text_4="Benchmark: In progress...",
            text_5="File: Nan",
            text_7="Current Task: DIST prediction...",
            text_8=print_progress_bar(0, len(self.eval_data)),
        )

        for i in range(len(self.eval_data)):
            """Predict"""
            (
                idx,
                coord_org,
                coords,
                nodes,
                target,
                output_idx,
                _,
            ) = self.eval_data.__getitem__(i)
            target = [t.cpu().detach().numpy() for t in target]
            output_idx = [o.cpu().detach().numpy() for o in output_idx]

            self.tardis_progress(
                title=self.title,
                text_1=f"Running point cloud segmentation benchmark on "
                f"{self.data_set}",
                text_4="Benchmark: In progress...",
                text_5=f"File: {idx}",
                text_7=self._update_metric_pg(),
                text_8="Current Task: DIST prediction...",
                text_9=print_progress_bar(i, len(self.eval_data)),
            )

            graphs = []
            for edge, graph, node in zip(coords, target, nodes):
                if self.rgb:
                    input = self._predict(edge[None, :], node[None, :])
                else:
                    input = self._predict(edge[None, :])
                graphs.append(input)

                """Benchmark Graph"""
                self._benchmark_graph(input, graph.astype(np.uint8))

            self.tardis_progress(
                title=self.title,
                text_1=f"Running point cloud segmentation benchmark on "
                f"{self.data_set}",
                text_4="Benchmark: In progress...",
                text_5=f"File: {idx}",
                text_7=self._update_metric_pg(),
                text_8="Current Task: DIST prediction...",
                text_9=print_progress_bar(i, len(self.eval_data)),
            )

            """Segment graphs"""
            self._benchmark_IS(graphs, coord_org, output_idx)

            self.tardis_progress(
                title=self.title,
                text_1=f"Running point cloud segmentation benchmark on "
                f"{self.data_set}",
                text_4="Benchmark: In progress...",
                text_5=f"File: {idx}",
                text_7=self._update_metric_pg(),
                text_8="Current Task: DIST prediction...",
                text_9=print_progress_bar(i, len(self.eval_data)),
            )

        return self._output_metric()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import time

import click
from tardis_em._version import version
from tardis_em.utils.logo import TardisLogo
from tardis_em.tardis_helper.helper_func import tardis_helper

ota = ""


@click.command()
@click.option(
    "-f",
    "--func",
    default=None,
    type=click.Choice(
        [
            "csv_am",
            "am_csv",
        ]
    ),
    help="Function name.",
    show_default=True,
)
@click.option(
    "-dir",
    "--dir_",
    default=None,
    type=str,
    help="Directory to files.",
    show_default=True,
)
@click.option(
    "-px",
    "--pixel_size",
    default=None,
    type=float,
    help="Optional pixel size value for image/coordinates conversion.",
    show_default=True,
)
@click.version_option(version=version)
def main(func=None, dir_=None, pixel_size=None):
    main_logo = TardisLogo()

    # Check if PyTorch was installed correctly with GPU support
    import torch

    res = torch.cuda.is_available()
    if not res:
        main_logo(
            title=f"| Transforms And Rapid Dimensionless Instance Segmentation | {ota}",
            text_0="WELCOME to TARDIS!",
            text_1="TARDIS detected no GPU support :(",
            text_3="Do not panic! Please uninstall pytorch,",
            text_4="and follow official instruction from https://pytorch.org",
            text_7="robert.kiewisz@gmail.com | tbepler@nysbc.org",
            text_8="Join Slack community: https://tardis-em.slack.com",
        )
        time.sleep(10)

    if func is not None and dir_ is not None:
        tardis_helper(func, dir_, pixel_size)
    else:
        main_logo(
            title=f"| Transforms And Rapid Dimensionless Instance Segmentation | {ota}",
            text_0="WELCOME to TARDIS!",
            text_1="TARDIS is fully automatic segmentation software no need for model training!",
            text_3="Contact developers if segmentation of your organelle is not supported! ",
            text_4="robert.kiewisz@gmail.com | tbepler@nysbc.org",
            text_5="Join Slack community: https://tardis-em.slack.com",
            text_7="FUNCTIONALITY:",
            text_8="To predict microtubule and filament instances:",
            text_9="  tardis_mt --help",
            text_10="To predict 3D membrane semantic and instances:",
            text_11=" tardis_mem --help |OR| tardis_mem2d --help",
        )


if __name__ == "__main__":
    main()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from math import sqrt
from typing import Optional

import torch
import torch.nn as nn

from tardis_em.utils.errors import TardisError


class GeLU(nn.Module):
    """
    Applies the Gaussian Error Linear Unit (GeLU) activation function.

    The GeLU activation function is a smooth approximation to the ReLU activation function.
    This implementation provides an optional `tanh` parameter that can be used to scale the
    input argument to the standard mathematical error function (erf). It is primarily used
    in neural networks to introduce non-linearity.
    """

    def __init__(self, tanh: Optional[float] = None):
        """
        Represents the Gaussian Error Linear Unit (GeLU) function used as an activation function
        in machine learning models. This implementation contains an optional tangent hyperbolic
        approximation factor that alters the behavior of the GeLU computation.

        :param tanh: Optional tangent hyperbolic approximation factor. If provided, its square
            root is calculated and stored as an attribute.
        :type tanh: float or None
        """
        super(GeLU, self).__init__()

        if tanh is not None:
            self.tanh = sqrt(tanh)
        else:
            self.tanh = 1.4142135623730951

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Applies a transformation to the input tensor using the hyperbolic tangent error function,
        scaling the result accordingly. This method is commonly utilized in machine learning models
        and functions by smoothing input data values into a specified scaling range.

        :param x: Input tensor on which the transformation is applied.
        :type x: torch.Tensor

        :return: Transformed tensor after applying the scaled error function operation.
        :rtype: torch.Tensor
        """
        return x * 0.5 * (1.0 + torch.erf(x / self.tanh))


def convolution(
    in_ch: int,
    out_ch: int,
    components: str,
    kernel: int or tuple,
    padding: int or tuple,
    num_group=None,
) -> list:
    """
    Builds a neural network block by assembling components specified in the given
    string. This function enables the construction of customizable CNN layers
    based on the input parameters and component configurations. The components
    can consist of various operations such as convolutions (2D or 3D), normalization
    techniques (GroupNorm, BatchNorm), and activation functions (ReLU, LeakyReLU,
    GeLU, PReLU). These components are added sequentially to the module list based
    on their order in the `components` string.

    :param in_ch: Number of input channels for the convolutional layers.
    :type in_ch: int
    :param out_ch: Number of output channels for the convolutional layers.
    :type out_ch: int
    :param components: A string that specifies the sequence of operations for the
                       CNN block. Each character or group represents a specific
                       operation such as convolution, normalization, or activation.
    :type components: str
    :param kernel: Kernel size for the convolution. Can be an integer or a tuple
                   representing the size.
    :type kernel: int or tuple
    :param padding: Padding size for the convolution. Can be an integer or a tuple
                    representing the size.
    :type padding: int or tuple
    :param num_group: Number of groups for Group Normalization. Required if
                      GroupNorm is included in the components. Defaults to None.
    :type num_group: int, optional

    :return: A list of tuples representing the layers in the CNN block.
             Each tuple contains a layer name and its corresponding module.
    :rtype: list
    """
    modules = []
    conv = False

    """Build CNN block"""
    for letter in components:
        """Add Convolution"""
        if "c" in letter:
            conv = True

            # Add 3DConv
            if "3" in components:
                if "g" in components or "b" in components:
                    modules.append(
                        (
                            "Conv3D",
                            nn.Conv3d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                                bias=False,
                            ),
                        )
                    )
                else:
                    modules.append(
                        (
                            "Conv3D",
                            nn.Conv3d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                            ),
                        )
                    )
            # Add 2DConv
            if "2" in components:
                if "g" in components or "b" in components:
                    modules.append(
                        (
                            "Conv2D",
                            nn.Conv2d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                                bias=False,
                            ),
                        )
                    )
                else:
                    modules.append(
                        (
                            "Conv2D",
                            nn.Conv2d(
                                in_channels=in_ch,
                                out_channels=out_ch,
                                kernel_size=kernel,
                                padding=padding,
                            ),
                        )
                    )
        """Add GroupNorm"""
        if "g" == letter:
            if num_group is None:
                TardisError(
                    "142",
                    "tardis_em/cnn/model/convolution.py",
                    "Number of group is required if nn.GroupNorm is used.",
                )

            if num_group > in_ch:
                num_group = 1
            if conv:
                modules.append(
                    (
                        "GroupNorm1",
                        nn.GroupNorm(num_groups=num_group, num_channels=out_ch),
                    )
                )
            else:
                modules.append(
                    (
                        "GroupNorm2",
                        nn.GroupNorm(num_groups=num_group, num_channels=in_ch),
                    )
                )
        """Add BatchNorm"""
        if "b" == letter:
            if "3" in components:
                if conv:
                    modules.append(("BatchNorm1", nn.BatchNorm3d(out_ch)))
                else:
                    modules.append(("BatchNorm2", nn.BatchNorm3d(in_ch)))
            if "2" in components:
                if conv:
                    modules.append(("BatchNorm1", nn.BatchNorm2d(out_ch)))
                else:
                    modules.append(("BatchNorm2", nn.BatchNorm2d(in_ch)))
        """Add ReLu"""
        if "r" in letter:
            modules.append(("ReLu", nn.ReLU()))
        """Add LeakyReLu"""
        if "l" in letter:
            modules.append(("LeakyReLu", nn.LeakyReLU(negative_slope=0.1)))
        """Add GeLu"""
        if "e" in letter:
            modules.append(("GeLU", GeLU()))
        if "p" in letter:
            modules.append(("PReLU", nn.PReLU(num_parameters=out_ch)))

    return modules


class SingleConvolution(nn.Sequential):
    """
    Represents a sequential layer that contains a single convolution operation.

    This class is a part of a neural network building process, specifically for
    performing either 2D or 3D convolutional operations. It inherits from
    `nn.Sequential` to combine multiple convolutional modules into a sequence
    based on the specified dimensionality of the operation. It dynamically
    constructs convolutional layers considering the input parameters.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        components: str,
        kernel: int or tuple,
        padding: int or tuple,
        num_group=None,
        block_type="any",
    ):
        """
        Represents a single convolution operation, supporting 2D or 3D convolution
        based on the input arguments. The class dynamically constructs and assigns
        the convolution modules during initialization for efficient and modular
        integration.

        :param in_ch: Number of input channels for the convolution operation.
        :type in_ch: int
        :param out_ch: Number of output channels from the convolution operation.
        :type out_ch: int
        :param components: Specification for the components to be utilized,
            determining if 2D or 3D convolution is applied. It should include
            '2' for 2D and '3' for 3D components processing.
        :type components: str
        :param kernel: Size of the convolution kernel. Accepts either an int or a
            tuple reflecting the dimensionality of the kernel.
        :type kernel: int or tuple
        :param padding: Padding value applied to the input, configurable by
            integer or tuple for kernel dimension matching.
        :type padding: int or tuple
        :param num_group: Optional grouping parameter for grouped convolution,
            defaults to None implying standard convolution behavior.
        :type num_group: int, optional
        :param block_type: An optional string parameter to define the specific
            block type functionality. Defaults to "any".
        :type block_type: str, optional
        """
        super(SingleConvolution, self).__init__()

        """Build single Conv3D"""
        if "3" in components:
            conv3d = convolution(
                in_ch=in_ch,
                out_ch=out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            )

            for name, module in conv3d:
                self.add_module(name, module)

        """Build single Conv2D"""
        if "2" in components:
            conv2d = convolution(
                in_ch=in_ch,
                out_ch=out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            )

            for name, module in conv2d:
                self.add_module(name, module)


class DoubleConvolution(nn.Sequential):
    """
    Implements a double convolutional block feature in a neural network.

    This class is a specialized neural network module implementing a double
    convolutional operation, designed either as an encoder block for reducing
    spatial resolution while increasing the number of features or as a decoder
    block for recovering spatial resolution with retained channel properties.
    It supports two block types - "encoder" and "decoder", leveraging the
    flexibility of defining kernel size, padding, and customizable components.
    The concept of group convolutions is enabled for enhanced modularity via
    `num_group`.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        block_type: str,
        kernel: int or tuple,
        padding: int or tuple,
        components="cgr",
        num_group=None,
    ):
        """
        This class initializes a double convolution block by chaining two convolutional layers
        using the specified input and output channel characteristics, block type, kernel properties,
        and other optional configurations. It dynamically computes the input and output channels
        for each convolution layer based on the block type.

        :param in_ch: The number of input channels for the convolution block.
        :param out_ch: The number of output channels for the convolution block.
        :param block_type: Specifies the type of block, either "encoder" or "decoder".
        :param kernel: The kernel size(s) for the convolution layers. Can be an integer or a tuple.
        :param padding: The padding size(s) for the convolution layers. Can be an integer or a tuple.
        :param components: Components to include in the convolution (default: "cgr").
        :param num_group: The number of groups for grouped convolutions (optional).

        :raises TardisError: Raised if the block type is not "encoder" or "decoder".
        """
        super(DoubleConvolution, self).__init__()

        # Define in and out channels for 1st and 2nd convolutions
        if block_type not in ["encoder", "decoder"]:
            TardisError(
                "143",
                "tardis_em/cnn/model/convolution.py",
                'Only "encoder", decoder block type is supported.',
            )

        """Calculate in and out channels for double convolution"""
        if block_type == "encoder":
            conv1_in_ch, conv1_out_ch = in_ch, out_ch // 2
            conv2_in_ch, conv2_out_ch = conv1_out_ch, out_ch
        else:
            conv1_in_ch, conv1_out_ch = in_ch, out_ch
            conv2_in_ch, conv2_out_ch = out_ch, out_ch

        """Build double convolution"""
        self.add_module(
            "DoubleConv1",
            SingleConvolution(
                in_ch=conv1_in_ch,
                out_ch=conv1_out_ch,
                kernel=kernel,
                padding=padding,
                components=components,
                num_group=num_group,
            ),
        )

        self.add_module(
            "DoubleConv2",
            SingleConvolution(
                in_ch=conv2_in_ch,
                out_ch=conv2_out_ch,
                components=components,
                kernel=kernel,
                padding=padding,
                num_group=num_group,
            ),
        )


class RecurrentDoubleConvolution(nn.Module):
    """
    Defines the RecurrentDoubleConvolution class, which implements a customized
    recurrent double convolution block designed for encoder and decoder
    operations in convolutional neural networks (CNN). The block consists of
    three consecutive single convolution operations with different configurations.
    It incorporates residual connections and supports various non-linearity
    components such as LeakyReLU, ReLU, and PReLU, making it suitable for
    flexible deep learning architectures.
    """

    def __init__(
        self,
        in_ch: int,
        out_ch: int,
        block_type: str,
        kernel: int or tuple,
        padding: int or tuple,
        components="cgr",
        num_group=None,
    ):
        """
        Represents a recurrent double convolution block used in constructing
        neural network architectures. This block can be configured as either
        an encoder or decoder block, transforming input feature maps into
        output feature maps through a sequence of convolutions and optional
        non-linearity.

        The structure adapts based on the block type to maintain flexibility
        for various deep learning tasks, such as image reconstruction or
        semantic segmentation.

        :param in_ch: Number of input channels.
        :type in_ch: int
        :param out_ch: Number of output channels.
        :type out_ch: int
        :param block_type: Specifies the type of convolution block ('encoder' or 'decoder').
        :type block_type: str
        :param kernel: Kernel size for the convolutional layers.
        :type kernel: int or tuple
        :param padding: Padding value for the convolutional layers.
        :type padding: int or tuple
        :param components: Components for the convolution operations ('c', 'g', 'r' or their combinations).
                           Defaults to "cgr".
        :type components: str
        :param num_group: Number of groups for group convolution. Defaults to None.
        :type num_group: int, optional
        :raises ValueError: If block_type is not 'encoder' or 'decoder'.
        """
        super(RecurrentDoubleConvolution, self).__init__()

        # Define in and out channels for 1st and 2nd convolutions
        if block_type not in ["encoder", "decoder"]:
            TardisError(
                "143",
                "tardis_em/cnn/model/convolution.py",
                'Only "encoder", decoder block type is supported.',
            )

        """Calculate in and out channels for double convolution"""
        if block_type == "encoder":
            conv1_in_ch, conv1_out_ch = in_ch, out_ch // 2
            conv2_in_ch, conv2_out_ch = conv1_out_ch, out_ch
        else:
            conv1_in_ch, conv1_out_ch = in_ch, out_ch
            conv2_in_ch, conv2_out_ch = out_ch, out_ch

        """Build RCNN block"""
        self.conv1 = SingleConvolution(
            in_ch=conv1_in_ch,
            out_ch=conv1_out_ch,
            kernel=kernel,
            padding=padding,
            components=components,
            num_group=num_group,
        )

        self.conv2 = SingleConvolution(
            in_ch=conv2_in_ch,
            out_ch=conv2_out_ch,
            components=components,
            kernel=kernel,
            padding=padding,
            num_group=num_group,
        )

        self.conv3 = SingleConvolution(
            in_ch=conv2_out_ch,
            out_ch=conv2_out_ch,
            components="c",
            kernel=kernel,
            padding=padding,
            num_group=num_group,
        )
        if "l" in components:
            self.non_linearity = nn.LeakyReLU(negative_slope=0.1, inplace=True)
        elif "r" in components:
            self.non_linearity = nn.ReLU(inplace=True)
        elif "p" in components:
            self.non_linearity = nn.PReLU()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Processes input tensor through multiple convolutional layers and applies a
        non-linearity function. Implements a residual connection between intermediate
        convolutions and the final output computation.

        :param x: Input tensor to process through the defined convolutional layers.
        :type x: torch.Tensor

        :return: Transformed tensor after applying convolutional layers, residual
            connection, and non-linearity.
        :rtype: torch.Tensor
        """
        out = self.conv1(x)
        out = self.conv2(out)
        residual = out

        out = self.conv3(out)

        out += residual
        out = self.non_linearity(out)

        return out



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import numpy as np
from sklearn.metrics import f1_score, precision_score, recall_score
from tqdm.contrib import tzip


def benchmark_cnn(logits: list, targets: list, reduce="mean", tqdm_pb=True):
    all_precisions = []
    all_recall = []
    all_f1 = []

    all_AP = []
    all_AP90 = []

    # Validation scores
    if tqdm_pb:
        zip_l = tzip(logits, targets)
    else:
        zip_l = zip(logits, targets)

    id_ = 0
    for logit, target in zip_l:
        print(id_)
        logit = logit.flatten()
        logit_th = np.where(logit > 0.5, 1, 0).astype(np.uint8).flatten()
        target = np.where(target > 0, 1, 0).astype(np.uint8).flatten()

        all_precisions.append(precision_score(target, logit_th))
        all_recall.append(recall_score(target, logit_th))
        all_f1.append(f1_score(target, logit_th))
        print(all_precisions, all_recall, all_f1)

        # Average precision scores
        AP_score, _, _, AP90 = AP(
            logits=[logit], targets=[target], AP90=True, reduce="mean"
        )
        all_AP.append(AP_score)
        all_AP90.append(AP90)
        print(all_AP, all_AP90)
    if reduce == "none":
        return all_precisions, all_recall, all_f1, all_AP, all_AP90
    else:
        assert reduce == "mean"
        return np.mean([all_precisions, all_recall, all_f1, all_AP, all_AP90], axis=1)


def confusion_matrix(logits, targets):
    confusion_vector = logits / targets

    tp = np.sum(confusion_vector == 1)
    fp = np.sum(confusion_vector == float("inf"))
    fn = np.sum(confusion_vector == 0)

    precision = tp / (tp + fp + 1e-16)
    rec = tp / (tp + fn + 1e-16)

    return precision, rec


def call_precision_at_recall(scores: np.ndarray, y: np.ndarray):
    all_th = np.linspace(0, 1, 1000)

    total_p = y.sum()
    scores = np.clip(scores, 0, 1)

    order = np.argsort(scores)
    scores = scores[order]
    y = y[order]

    tp_at_th = np.cumsum(y[::-1])[::-1]
    pp_at_th = np.arange(len(scores), 0, -1)

    indices = np.searchsorted(scores, all_th, side="right")
    # indices[indices == len(tp_at_th)] = len(tp_at_th) - 1
    indices = np.minimum(indices, len(tp_at_th) - 1)

    tp_at_th = tp_at_th[indices]
    pp_at_th = pp_at_th[indices]

    precision = tp_at_th / pp_at_th
    recall = tp_at_th / total_p

    precision[pp_at_th == 0] = 1

    return precision, recall, all_th


def precision_at_recall(precision, recall, all_th, level):
    order = np.argsort(recall)
    recall = recall[order]
    precision = precision[order]
    all_th = all_th[order]

    # mask_1 = recall <= level
    # mask_2 = recall >= level
    #
    # i = np.where(mask_1)[0][-1]
    # j = np.where(mask_2)[0][0]

    # Find the indices of the recall values just below and just above the desired level
    i = np.searchsorted(recall, level, side="right") - 1
    j = np.searchsorted(recall, level, side="left")

    if i < 0:
        i = 0
    if j >= len(recall):
        j = len(recall) - 1

    ri = recall[i]
    rj = recall[j]

    pi = precision[i]
    pj = precision[j]

    ti = all_th[i]
    tj = all_th[j]

    p = pi + (level - ri) / (rj - ri) * (pj - pi)
    th = ti + (level - ri) / (rj - ri) * (tj - ti)

    return p, th


def AP(logits: list, targets: list, AP90=False, reduce="none"):
    assert len(logits) == len(
        targets
    ), "Length of logits and targets should be the same."
    assert reduce in ["none", "mean", "median"]

    all_precisions = np.zeros_like(np.linspace(0, 1, 1000))
    all_recalls = np.zeros_like(np.linspace(0, 1, 1000))
    ap90_scores = []

    for logit, target in zip(logits, targets):
        target = np.where(target > 0, 1, 0).astype(np.uint8).flatten()
        logit = logit.flatten()

        precision, recall, th = call_precision_at_recall(scores=logit, y=target)

        all_precisions += precision
        all_recalls += recall

        if AP90:
            ap90_score, th = precision_at_recall(precision, recall, th, 0.9)
            ap90_scores.append(ap90_score)

    all_precisions = all_precisions / len(logits)
    all_recalls = all_recalls / len(logits)
    if reduce == "none":
        if AP90:
            return all_precisions, all_recalls, th, ap90_scores
        return all_precisions, all_recalls, th
    elif reduce == "mean":
        all_precisions = np.mean(all_precisions)
        all_recalls = np.mean(all_recalls)
        ap90_scores = np.mean(ap90_scores)

        if AP90:
            return all_precisions, all_recalls, th, ap90_scores
        return all_precisions, all_recalls, th
    else:
        all_precisions = np.median(all_precisions)
        all_recalls = np.median(all_recalls)
        ap90_scores = np.median(ap90_scores)

        if AP90:
            return all_precisions, all_recalls, th, ap90_scores
        return all_precisions, all_recalls, th



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from typing import Optional

import torch

from tardis_em.cnn.cnn import build_cnn_network

structure = {
    "in_channel": 1,
    "out_channel": 1,
    "img_size": 64,
    "dropout": None,
    "conv_kernel": 3,
    "conv_padding": 1,
    "maxpool_kernel": 2,
    "num_group": 1,
    "num_conv_layers": 5,
    "conv_scaler": 32,
    "classification": False,
    "layer_components": "3gcl",
}


def unet_3d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "3gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="unet", structure=structure, img_size=image_size, prediction=False
    )


def unet_2d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "2gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="unet", structure=structure, img_size=image_size, prediction=False
    )


def resunet_3d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "3gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="resunet",
        structure=structure,
        img_size=image_size,
        prediction=False,
    )


def resunet_2d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "2gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="resunet",
        structure=structure,
        img_size=image_size,
        prediction=False,
    )


def unet3plus_3d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "3gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="unet3plus",
        structure=structure,
        img_size=image_size,
        prediction=False,
    )


def unet3plus_2d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "2gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="unet3plus",
        structure=structure,
        img_size=image_size,
        prediction=False,
    )


def fnet_3d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "3gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="fnet", structure=structure, img_size=image_size, prediction=False
    )


def fnet_2d(
    image_size: int, layer_num: int, conv_scaler: int, dropout: Optional[float] = None
):
    structure.update({"layer_components": "2gcl"})
    structure.update({"num_conv_layers": layer_num})
    structure.update({"conv_scaler": conv_scaler})
    structure.update({"dropout": dropout})
    return build_cnn_network(
        network_type="fnet", structure=structure, img_size=image_size, prediction=False
    )


class TestNetwork3D:
    image_sizes = [16, 32, 64, 96]
    conv_scaler = [4, 8, 16, 32]

    def test_unet3d(self):
        for i in self.image_sizes:
            for j in self.conv_scaler:
                # Batch x Channel x Z x Y x X
                input = torch.rand((1, 1, i, i, i))

                nn = unet_3d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)

                with torch.no_grad():
                    nn(input)

    def test_unet2d(self):
        for i in self.image_sizes:
            for j in self.conv_scaler:
                # Batch x Channel x Z x Y x X
                input = torch.rand((1, 1, i, i))

                nn = unet_2d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)

                with torch.no_grad():
                    nn(input)

    def test_resunet_3d(self):
        for i in self.image_sizes:
            for j in self.conv_scaler:
                # Batch x Channel x Z x Y x X
                input = torch.rand((1, 1, i, i, i))

                nn = resunet_3d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)

                with torch.no_grad():
                    nn(input)

    def test_resunet_2d(self):
        for i in self.image_sizes:
            for j in self.conv_scaler:
                # Batch x Channel x Z x Y x X
                input = torch.rand((1, 1, i, i))

                nn = resunet_2d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)

                with torch.no_grad():
                    nn(input)

    # def test_unet3plus_2d(self):
    #     for i in self.image_sizes:
    #         for j in self.conv_scaler:
    #             # Batch x Channel x Z x Y x X
    #             input = torch.rand((1, 1, i, i))
    #
    #             nn = unet3plus_2d(
    #                 image_size=i, layer_num=5, conv_scaler=j, dropout=None
    #             )
    #
    #             with torch.no_grad():
    #                 nn(input)
    #
    # def test_unet3plus_3d(self):
    #     for i in self.image_sizes:
    #         for j in self.conv_scaler:
    #             # Batch x Channel x Z x Y x X
    #             input = torch.rand((1, 1, i, i, i))
    #
    #             nn = unet3plus_3d(
    #                 image_size=i, layer_num=5, conv_scaler=j, dropout=None
    #             )
    #
    #             with torch.no_grad():
    #                 nn(input)

    # def test_fnet_2d(self):
    #     for i in self.image_sizes:
    #         for j in self.conv_scaler:
    #             # Batch x Channel x Z x Y x X
    #             input = torch.rand((1, 1, i, i))
    #
    #             nn = fnet_2d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)
    #
    #             with torch.no_grad():
    #                 nn(input)
    #
    # def test_fnet_3d(self):
    #     for i in self.image_sizes:
    #         for j in self.conv_scaler:
    #             # Batch x Channel x Z x Y x X
    #             input = torch.rand((1, 1, i, i, i))
    #
    #             nn = fnet_3d(image_size=i, layer_num=5, conv_scaler=j, dropout=None)
    #
    #             with torch.no_grad():
    #                 nn(input)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import os
from os import listdir
from os.path import join, splitext
from typing import Tuple

import numpy as np
import torch
from torch.utils.data import Dataset

from tardis_em.cnn.datasets.augmentation import preprocess
from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import load_image
from tardis_em.utils.normalization import MinMaxNormalize


class CNNDataset(Dataset):
    """
    Handles dataset creation and processing for Convolutional Neural Network (CNN) training and inference.

    This class manages the loading, processing, and formatting of image and mask data
    needed for training CNN models. It includes normalization, size adjustments,
    and optional transformations to prepare data for further model usage.
    """

    def __init__(
        self,
        img_dir: str,
        mask_dir: str,
        size=64,
        mask_suffix="_mask",
        transform=True,
        out_channels=1,
    ):
        """
        Represents a dataset class designed to accommodate image and mask directories for image
        processing tasks. This class initializes various properties including image and
        mask directories, normalization strategies, and options for transformation. It manages
        the internal identifiers for files in the directories and detects the file format
        based on sample files from the image directory.

        :param img_dir: Path to the directory containing the input images.
        :param mask_dir: Path to the directory containing the masks for the input images.
        :param size: Desired size of the images and masks after resizing. Defaults to 64.
        :param mask_suffix: Suffix added to the filenames of masks to associate them
            with corresponding input images. Defaults to "_mask".
        :param transform: Boolean value indicating whether to apply transformations
            to the images and masks. Defaults to True.
        :param out_channels: Number of output channels for processed data.
            Typically used in segmentation tasks. Defaults to 1.
        """
        self.img_dir = img_dir
        self.mask_dir = mask_dir
        self.size = size
        self.mask_suffix = mask_suffix
        self.transform = transform
        self.out_channels = out_channels
        self.minmax = MinMaxNormalize()

        self.ids = [
            splitext(file)[0] for file in listdir(img_dir) if not file.startswith(".")
        ]
        self.format = [
            splitext(file)[1]
            for file in listdir(img_dir)[:5]
            if not file.startswith(".")
        ][1]

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i: int) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Retrieves the image and corresponding label mask located at the specified index.
        The function performs the following steps:
        1. Identifies image and mask filenames using the given index.
        2. Loads the image and its associated mask.
        3. Validates their data types and ensures a proper binary range for the image.
        4. Preprocesses the image and mask into the desired format and size.
        5. Returns both as PyTorch tensors.

        :param i: Index of the desired image-mask pair within the dataset.
        :type i: int

        :return: Tuple containing a preprocessed image tensor and the corresponding mask tensor.
        :rtype: Tuple[torch.Tensor, torch.Tensor]
        """
        # Find next image and corresponding label mask image
        idx = self.ids[i]
        mask_file = os.path.join(self.mask_dir, str(idx) + "_mask" + self.format)
        img_file = os.path.join(self.img_dir, str(idx) + self.format)

        # Load image and corresponding label mask
        img, _ = load_image(img_file)
        mask, _ = load_image(mask_file)

        if mask.dtype != np.uint8:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                f"Mask should be of np.uint8 dtype but is {mask.dtype}!",
            )

        # Process image and mask
        img, mask = preprocess(
            image=img,
            mask=mask,
            size=self.size,
            transformation=self.transform,
            output_dim_mask=self.out_channels,
        )

        if img.dtype != np.float32 and mask.dtype != np.uint8:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                f"Mask {mask.dtype} and image  {img.dtype} has wrong dtype!",
            )
        if not img.min() >= -1 and not img.max() <= 1:
            TardisError(
                "147",
                "tardis_em/cnn/dataset/dataloader.py",
                "Image file is not binary!",
            )

        return torch.from_numpy(img.copy()).type(torch.float32), torch.from_numpy(
            mask.copy()
        ).type(torch.float32)


class PredictionDataset(Dataset):
    """
    Manages loading, processing, and serving of datasets for image prediction tasks.

    This class provides methods to load images from a specified directory, preprocess
    and format them into tensors compatible with convolutional neural networks. It also
    facilitates retrieval by a specific index. Used for predictive model input preparation.
    """

    def __init__(self, img_dir: str, out_channels=1):
        """
        This class initializes with the specified image directory path and output channel
        configuration. It lists all the files in the directory, excluding hidden files,
        and extracts their base names (without extension).

        :param img_dir: Path to the directory containing images. Represents the input
            directory from which image files' names will be retrieved and stored.
        :param out_channels: An integer representing the number of output channels to
            be used or processed. Default value is set to 1.
        """
        self.img_dir = img_dir
        self.out_channels = out_channels

        self.ids = [
            splitext(file)[0] for file in listdir(img_dir) if not file.startswith(".")
        ]

    def __len__(self):
        return len(self.ids)

    def __getitem__(self, i: int):
        """
        Access an item from the dataset by its index. This method retrieves an image
        given its index from the dataset, processes it, and returns the preprocessed
        image along with the corresponding index.

        :param i: An integer index of the image to retrieve from the dataset.
        :type i: int

        :return: A tuple containing the preprocessed image as a tensor and the
                 corresponding index of the image in the dataset.
        :rtype: Tuple[torch.Tensor, int]
        """
        idx = self.ids[i]
        img_file = join(self.img_dir, str(idx) + ".tif")

        # Load image
        img, _ = load_image(img_file)
        img = img.astype(np.float32)

        # Process image and mask
        img = preprocess(
            image=img,
            size=img.shape,
            transformation=False,
            output_dim_mask=self.out_channels,
        )

        return torch.from_numpy(img).type(torch.float32), idx



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from math import ceil
from os import mkdir
from os.path import isdir, join
from typing import Optional, Tuple

import numpy as np
from tifffile import tifffile as tif

from tardis_em.cnn.data_processing.scaling import scale_image
from tardis_em.utils.device import get_device
from tardis_em.utils.export_data import to_mrc
from tardis_em.utils.errors import TardisError


def trim_with_stride(
    image: np.ndarray,
    trim_size_xy: int,
    trim_size_z: int,
    output: str,
    image_counter: int,
    scale: list,
    clean_empty=True,
    keep_if=0.01,
    stride=25,
    mask: Optional[np.ndarray] = None,
    log=True,
    pixel_size=None,
    device=get_device("cpu"),
):
    """
    Trim a 2D/3D image with optional masking into smaller patches with a specific stride
    and size. This function is designed for preprocessing image data for machine learning
    tasks, ensuring that the image dimensions meet specified requirements or adjusting
    them when necessary.

    :param image:
        The input image array, which can be 2D or 3D, and include grayscale or RGB
        channels. The data type and shape may vary depending on the input.
    :param trim_size_xy:
        Desired size for trimming in the X and Y dimensions. The input value determines
        the width and height of each patch. If not provided, default values will be
        calculated.
    :param trim_size_z:
        Desired size for trimming in the Z dimension. The input value determines
        the depth of each patch in case of 3D data. If not provided, default values
        will be calculated.
    :param output:
        Directory path where the trimmed patches will be saved. Output images and masks
        are stored in subdirectories.
    :param image_counter:
        Counter used for naming the output files. This ensures unique filenames during
        the trimming process.
    :param scale:
        A scaling factor list that adjusts the image size before trimming is performed.
    :param clean_empty:
        When set to True, automatic filtering is applied to exclude patches considered
        empty based on a specific condition.
    :param keep_if:
        Minimum threshold for keeping patches. The condition ensures that patches
        with non-empty values above this percentage are retained.
    :param stride:
        The stride size for trimming patches. Stride determines the overlap between
        adjacent patches when slicing the image.
    :param mask:
        Optional mask array corresponding to the input image. Used for segmentation
        or region-specific analyses. The function scales and trims the mask alongside
        the image.
    :param log:
        Flag to enable or disable logging of the process; logs errors and important
        actions during execution when set to True.
    :param pixel_size:
        Indicates whether the output should be saved in '.mrc' (specific file format)
        or '.tif' (common image format). Leave None for default behavior.
    :param device:
        Specifies the computational device (CPU, GPU, etc.) to be used if scaling
        operations employ device-specific optimizations.

    :return:
        Returns None. Side effect: Saves the image patches and corresponding mask
        patches (when provided) to the specified output directory. Each patch is
        named according to its positional indices to maintain unique identification.
    """
    img_dtype = np.float32

    if not isdir(join(output, "imgs")):
        mkdir(join(output, "imgs"))
    if not isdir(join(output, "masks")) and mask is not None:
        mkdir(join(output, "masks"))

    if mask is not None:
        mask_dtype = np.uint8
        image, mask, dim = scale_image(
            image=image,
            mask=mask,
            scale=tuple(scale),
            device=device,
        )
        mask = np.where(mask > 0, 1, 0).astype(np.uint8)

        if image.shape != mask.shape:
            TardisError(
                "111",
                "tardis_em/cnn/data_processing.md/trim.py",
                f"Image {image.shape} has different shape from mask {mask.shape}",
            )
    else:
        image, dim = scale_image(image=image, scale=scale, device=device)

    if img_dtype != image.dtype:
        TardisError(
            "111",
            "tardis_em/cnn/data_processing.md/trim.py",
            f"Image {img_dtype} has different dtype after interpolation {image.dtype}",
        )

    nz, ny, nx, nc = 0, 0, 0, None
    if image.ndim == 4 and dim == 3:  # 3D RGB
        nz, ny, nx, nc = image.shape
    elif image.ndim == 3 and dim == 3:  # 2D RGB
        ny, nx, nc = image.shape
        nz = 0  # 2D
    elif image.ndim == 3 and dim == 1:  # 3D gray
        nz, ny, nx = image.shape
        nc = None  # Gray
    elif image.ndim == 2:  # 2D gray
        ny, nx = image.shape
        nc = None  # 2D
        nz = 0  # Gray

    if stride is None:
        TardisError(
            "112",
            "tardis_em/cnn/data_processing.md",
            "Trim sizes or stride has to be indicated!",
        )

    """Calculate number of patches and stride for xyz"""
    x, y, z = ceil(nx / trim_size_xy), ceil(ny / trim_size_xy), ceil(nz / trim_size_z)

    if nz > 0:
        x_pad, y_pad, z_pad = (
            (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx,
            (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny,
            (trim_size_z + (trim_size_z - stride) * (z - 1)) - nz,
        )
    else:
        x_pad, y_pad, z_pad = (
            (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx,
            (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny,
            0,
        )

    """Adapt number of patches or patch size for trimming"""
    if trim_size_xy is not None or trim_size_z is not None:
        while x_pad < 0:
            x += 1
            x_pad += trim_size_xy - stride
        while y_pad < 0:
            y += 1
            y_pad += trim_size_xy - stride
        if nz > 0:
            while z_pad < 0:
                z += 1
                z_pad += trim_size_z - stride
    else:
        while x_pad <= 0 or y_pad <= 0:
            trim_size_xy += 1
            x_pad = (trim_size_xy + ((trim_size_xy - stride) * (x - 1))) - nx
            y_pad = (trim_size_xy + ((trim_size_xy - stride) * (y - 1))) - ny

        if nz > 0:
            while z_pad < 0:
                trim_size_z += 1
                z_pad = (trim_size_z + ((trim_size_z - stride) * (z - 1))) - nz
        else:
            z_pad = 0

    """Expand image of a patch"""
    if nz > 0:
        if nc is not None:
            image_padded = np.pad(image, [(0, z_pad), (0, y_pad), (0, x_pad), (0, 0)])
        else:
            image_padded = np.pad(image, [(0, z_pad), (0, y_pad), (0, x_pad)])

        if mask is not None:
            mask_padded = np.pad(mask, [(0, z_pad), (0, y_pad), (0, x_pad)])
    else:
        if nc is not None:
            image_padded = np.pad(image, [(0, y_pad), (0, x_pad), (0, 0)])
        else:
            image_padded = np.pad(image, [(0, y_pad), (0, x_pad)])

            if mask is not None:
                mask_padded = np.pad(mask, [(0, y_pad), (0, x_pad)])

    """Trim image and mask with stride"""
    z_start, z_stop = 0 - (trim_size_z - stride), 0
    count = len(range(x)) * len(range(y))
    if len(range(z)) > 0:
        count = count * len(range(z))
    print(f"Found {count} patches...")

    count_save = 0
    if z == 0:
        z = 1

    for i in range(z):
        z_start = z_start + trim_size_z - stride
        z_stop = z_start + trim_size_z
        y_start, y_stop = 0 - (trim_size_xy - stride), 0

        for j in range(y):
            y_start = y_start + trim_size_xy - stride
            y_stop = y_start + trim_size_xy
            x_start, x_stop = 0 - (trim_size_xy - stride), 0

            for k in range(x):
                x_start = x_start + trim_size_xy - stride
                x_stop = x_start + trim_size_xy

                if pixel_size is None:
                    img_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}.tif")
                    mask_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}_mask.tif")
                else:
                    img_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}.mrc")
                    mask_name = str(f"{image_counter}_{i}_{j}_{k}_{stride}_mask.mrc")

                if nc is None:
                    if nz > 0:
                        trim_img = image_padded[
                            z_start:z_stop, y_start:y_stop, x_start:x_stop
                        ]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]
                    else:
                        trim_img = image_padded[y_start:y_stop, x_start:x_stop]
                        if mask is not None:
                            trim_mask = mask_padded[y_start:y_stop, x_start:x_stop]
                else:
                    if nz > 0:
                        trim_img = image_padded[
                            z_start:z_stop, y_start:y_stop, x_start:x_stop, :
                        ]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]
                    else:
                        trim_img = image_padded[y_start:y_stop, x_start:x_stop, :]
                        if mask is not None:
                            trim_mask = mask_padded[
                                z_start:z_stop, y_start:y_stop, x_start:x_stop
                            ]

                trim_img = np.array(trim_img, dtype=img_dtype)

                if clean_empty and mask is not None:
                    if np.sum(trim_mask) > keep_if:
                        count_save += 1

                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                            tif.imwrite(
                                join(output, "masks", mask_name),
                                np.array(trim_mask, dtype=mask_dtype),
                                shape=trim_mask.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))
                            to_mrc(
                                np.array(trim_mask, mask_dtype),
                                pixel_size,
                                join(output, "masks", mask_name),
                            )
                else:
                    count_save += 1
                    if mask is None:
                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))

                    else:
                        if pixel_size is None:
                            tif.imwrite(
                                join(output, "imgs", img_name),
                                trim_img,
                                shape=trim_img.shape,
                            )
                            tif.imwrite(
                                join(output, "masks", mask_name),
                                np.array(trim_mask, dtype=mask_dtype),
                                shape=trim_mask.shape,
                            )
                        else:
                            to_mrc(trim_img, pixel_size, join(output, "imgs", img_name))
                            to_mrc(
                                np.array(trim_mask, mask_dtype),
                                pixel_size,
                                join(output, "masks", mask_name),
                            )

    if log:
        return [count, count_save]


def trim_label_mask(
    points: np.ndarray, image: np.ndarray, label_mask: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Trim the label mask, image, and adjust the point coordinates based on the minimum and
    maximum boundaries in the provided points. The function calculates the bounding box
    using the points array, crops the image and label mask within this bounding box,
    and shifts the points coordinates accordingly.

    :param points: A numpy array of shape (N, 3) representing the coordinates of points.
    :param image: A 3D numpy array representing the volumetric image.
    :param label_mask: A 3D numpy array representing the volumetric label mask.

    :return: A tuple containing the trimmed image array, the trimmed label_mask array,
        and the adjusted points array.
    :rtype: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """
    max_x, min_x = max(points[:, 0]), min(points[:, 0])
    max_y, min_y = max(points[:, 1]), min(points[:, 1])
    max_z, min_z = max(points[:, 2]), min(points[:, 2])

    if min_z < 0:
        min_z = 0

    image_trim = image[
        int(min_z) : int(max_z), int(min_y) : int(max_y), int(min_x) : int(max_x)
    ]

    label_mask_trim = label_mask[
        int(min_z) : int(max_z), int(min_y) : int(max_y), int(min_x) : int(max_x)
    ]

    points[:, 0] = points[:, 0] - min_x
    points[:, 1] = points[:, 1] - min_y
    points[:, 2] = points[:, 2] - min_z

    return image_trim, label_mask_trim, points



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from os import getcwd, listdir, mkdir
from os.path import isdir, join
from shutil import rmtree
from typing import Optional, Union

import click
import torch
from torch.utils.data import DataLoader

from tardis_em.cnn.datasets.build_dataset import build_train_dataset
from tardis_em.cnn.datasets.dataloader import CNNDataset
from tardis_em.cnn.train import train_cnn
from tardis_em.utils.dataset import build_test_dataset
from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.logo import TardisLogo
from tardis_em.utils.setup_envir import check_dir
from tardis_em._version import version


@click.command()
@click.option(
    "-dir",
    "--path",
    default=getcwd(),
    type=str,
    help="Directory with train, test folder or folder with dataset "
    "to be used for training.",
    show_default=True,
)
@click.option(
    "-ps",
    "--patch_size",
    default=64,
    type=int,
    help="Image size used for prediction.",
    show_default=True,
)
@click.option(
    "-px",
    "--pixel_size",
    default=None,
    type=float,
    help="Pixel size in [A] to which all images are resize.",
    show_default=True,
)
@click.option(
    "-ms",
    "--mask_size",
    default=150,
    type=int,
    help="Size of drawn mask in A. If you are using .csv files as labels.",
    show_default=True,
)
@click.option(
    "-cnn",
    "--cnn_type",
    default="unet",
    type=click.Choice(
        ["unet", "resunet", "unet3plus", "big_unet", "fnet", "fnet_attn"]
    ),
    help="Type of NN used for training.",
    show_default=True,
)
@click.option(
    "-co",
    "--cnn_out_channel",
    default=1,
    type=int,
    help="Number of output channels for the NN.",
    show_default=True,
)
@click.option(
    "-b",
    "--training_batch_size",
    default=25,
    type=int,
    help="Batch size.",
    show_default=True,
)
@click.option(
    "-cl",
    "--cnn_layers",
    default=5,
    type=int,
    help="Number of convolution layer for NN.",
    show_default=True,
)
@click.option(
    "-cm",
    "--cnn_scaler",
    default=32,
    type=int,
    help="Convolution multiplayer for CNN layers.",
    show_default=True,
)
@click.option(
    "-cs",
    "--cnn_structure",
    default="3gcl",
    type=str,
    help="Define structure of the convolution layer."
    "2 or 3 - dimension in 2D or 3D"
    "c - convolution"
    "g - group normalization"
    "b - batch normalization"
    "r - ReLU"
    "l - LeakyReLU"
    "e - GeLu"
    "p - PReLu",
    show_default=True,
)
@click.option(
    "-ck",
    "--conv_kernel",
    default=3,
    type=int,
    help="Kernel size for 2D or 3D convolution.",
    show_default=True,
)
@click.option(
    "-cp",
    "--conv_padding",
    default=1,
    type=int,
    help="Padding size for convolution.",
    show_default=True,
)
@click.option(
    "-cmpk",
    "--pool_kernel",
    default=2,
    type=int,
    help="Max_pooling kernel.",
    show_default=True,
)
@click.option(
    "-l",
    "--cnn_loss",
    default="BCELoss",
    type=str,
    help="Loss function use for training.",
    show_default=True,
)
@click.option(
    "-lr",
    "--loss_lr_rate",
    default=0.0005,
    type=float,
    help="Learning rate for NN.",
    show_default=True,
)
@click.option(
    "-lrs",
    "--lr_rate_schedule",
    default=False,
    type=bool,
    help="If True learning rate scheduler is used.",
    show_default=True,
)
@click.option(
    "-dv",
    "--device",
    default=0,
    type=str,
    help="Define which device use for training: "
    "gpu: Use ID 0 gpus"
    "cpu: Usa CPU"
    "mps: Apple silicon"
    "0-9 - specified gpu device id to use",
    show_default=True,
)
@click.option(
    "-w",
    "--warmup",
    default=100,
    type=int,
    help="Number of warmup steps.",
    show_default=True,
)
@click.option(
    "-e",
    "--epochs",
    default=10000,
    type=int,
    help="Number of epoches.",
    show_default=True,
)
@click.option(
    "-es",
    "--early_stop",
    default=1000,
    type=int,
    help="Number of epoches without improvement after which early stop "
    "is initiated. Default should is 10% of the total number of epochs.",
    show_default=True,
)
@click.option(
    "-cch",
    "--cnn_checkpoint",
    default=None,
    type=str,
    help="If indicated, dir to training checkpoint to reinitialized training.",
    show_default=True,
)
@click.option(
    "-dp",
    "--dropout_rate",
    default=0.5,
    type=float,
    help="If indicated, value of dropout for CNN.",
    show_default=True,
)
@click.option("-test_click", "--test_click", default=False, hidden=True)
@click.version_option(version=version)
def main(
    path: str,
    patch_size: int,
    pixel_size: Union[float, None],
    mask_size: int,
    cnn_type: str,
    cnn_out_channel: int,
    training_batch_size: int,
    cnn_layers: int,
    cnn_scaler: int,
    conv_kernel: int,
    conv_padding: int,
    pool_kernel: int,
    cnn_structure: str,
    cnn_loss: str,
    loss_lr_rate: float,
    lr_rate_schedule: bool,
    device: str,
    warmup: int,
    epochs: int,
    early_stop: int,
    cnn_checkpoint: Optional[str] = None,
    dropout_rate: Optional[float] = None,
    test_click=False,
):
    """
    MAIN MODULE FOR TRAINING CNN UNET/RESUNET/UNET3PLUS MODELS
    """
    """Initialize TARDIS progress bar"""
    tardis_logo = TardisLogo()
    tardis_logo(title="CNN training module")

    """Set environment"""
    TRAIN_IMAGE_DIR = join(path, "train", "imgs")
    TRAIN_MASK_DIR = join(path, "train", "masks")
    TEST_IMAGE_DIR = join(path, "test", "imgs")
    TEST_MASK_DIR = join(path, "test", "masks")

    IMG_FORMAT = (".tif", ".am", ".mrc", ".rec", ".map")

    """Check if dir has train/test folder and if folder have compatible data"""
    DATASET_TEST = check_dir(
        dir_s=path,
        with_img=True,
        train_img=TRAIN_IMAGE_DIR,
        train_mask=TRAIN_MASK_DIR,
        img_format=IMG_FORMAT,
        test_img=TEST_IMAGE_DIR,
        test_mask=TEST_MASK_DIR,
        mask_format=(
            "_mask.am",
            ".CorrelationLines.am",
            "_mask.mrc",
            "_mask.tif",
            "_mask.csv",
        ),
    )

    """Optionally: Set-up environment if not existing"""
    if not DATASET_TEST:
        # Check and set-up environment
        if not len([f for f in listdir(path) if f.endswith(IMG_FORMAT)]) > 0:
            TardisError(
                "100",
                "tardis_em/train_cnn.py",
                "Indicated folder for training do not have any compatible "
                "data or one of the following folders: "
                "test/imgs; test/masks; train/imgs; train/masks",
            )

        if isdir(join(path, "train")):
            rmtree(join(path, "train"))

        mkdir(join(path, "train"))
        mkdir(TRAIN_IMAGE_DIR)
        mkdir(TRAIN_MASK_DIR)

        if isdir(join(path, "test")):
            rmtree(join(path, "test"))

        mkdir(join(path, "test"))
        mkdir(TEST_IMAGE_DIR)
        mkdir(TEST_MASK_DIR)

        # Build train and test dataset
        build_train_dataset(
            dataset_dir=path,
            circle_size=mask_size,
            resize_pixel_size=pixel_size,
            trim_xy=patch_size,
            trim_z=patch_size,
        )

        no_dataset = int(len([f for f in listdir(path) if f.endswith(IMG_FORMAT)]) / 2)
        build_test_dataset(dataset_dir=path, dataset_no=no_dataset)

    """Build training and test dataset 2D/3D"""
    if not test_click:
        train_DL = DataLoader(
            dataset=CNNDataset(
                img_dir=TRAIN_IMAGE_DIR,
                mask_dir=TRAIN_MASK_DIR,
                size=patch_size,
                out_channels=cnn_out_channel,
            ),
            batch_size=training_batch_size,
            shuffle=True,
            num_workers=8,
            pin_memory=True,
        )

        test_DL = DataLoader(
            dataset=CNNDataset(
                img_dir=TEST_IMAGE_DIR,
                mask_dir=TEST_MASK_DIR,
                size=patch_size,
                out_channels=cnn_out_channel,
            ),
            shuffle=True,
            num_workers=8,
            pin_memory=True,
        )

    if cnn_out_channel > 1:
        cnn_loss = "CELoss"

    if loss_lr_rate == 1.0:
        lr_rate_schedule = True

    """Get device"""
    device = get_device(device)

    """Model structure dictionary"""
    """Optionally: pre-load model structure from checkpoint"""
    model_dict = None
    if cnn_checkpoint is not None:
        save_train = torch.load(cnn_checkpoint, map_location=device, weights_only=False)

        if "model_struct_dict" in save_train.keys():
            model_dict = save_train["model_struct_dict"]
            globals().update(model_dict)
    if model_dict is None:
        model_dict = {
            "cnn_type": cnn_type,
            "classification": False,
            "in_channel": 1,
            "out_channel": cnn_out_channel,
            "img_size": patch_size,
            "dropout": dropout_rate,
            "num_conv_layers": cnn_layers,
            "conv_scaler": cnn_scaler,
            "conv_kernel": conv_kernel,
            "conv_padding": conv_padding,
            "maxpool_kernel": pool_kernel,
            "layer_components": cnn_structure,
            "attn_features": True if cnn_type == "fnet_attn" else False,
            "num_group": 8,
            "prediction": False,
        }

    """Run Training loop"""
    if not test_click:
        train_cnn(
            train_dataloader=train_DL,
            test_dataloader=test_DL,
            model_structure=model_dict,
            checkpoint=cnn_checkpoint,
            loss_function=cnn_loss,
            learning_rate=loss_lr_rate,
            learning_rate_scheduler=lr_rate_schedule,
            early_stop_rate=early_stop,
            device=device,
            warmup=warmup,
            epochs=epochs,
        )
    else:
        rmtree(join(path, "train"))
        rmtree(join(path, "test"))


if __name__ == "__main__":
    main()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from os.path import join

import numpy as np
import torch

from tardis_em.cnn.datasets.augmentation import preprocess
from tardis_em.cnn.datasets.dataloader import (
    CNNDataset,
    PredictionDataset,
)
from tardis_em.utils.normalization import (
    MinMaxNormalize,
    RescaleNormalize,
    SimpleNormalize,
)


def test_dataloader():
    dir_ = "./tests/test_data/data_loader/cnn/test/"
    dataset = CNNDataset(img_dir=join(dir_, "imgs"), mask_dir=join(dir_, "masks"))
    assert len(dataset) == 2

    img, mask = dataset.__getitem__(0)

    assert img.shape == (1, 64, 64, 64) and mask.shape == (1, 64, 64, 64)
    assert img.min() >= 0 and img.max() <= 255
    assert torch.sum(img) != 0

    dataset = PredictionDataset(img_dir=join(dir_, "imgs"))
    assert len(dataset) == 2

    img, idx = dataset.__getitem__(0)

    assert img.shape == (1, 64, 64, 64)
    assert img.min() >= -1 and img.max() <= 255
    assert torch.sum(img) != 0
    assert isinstance(idx, str) is True


def test_normalization():
    img = np.random.rand(64, 64, 64) * 255
    img = img.astype(np.uint8)

    s_norm = SimpleNormalize()
    mm_norm = MinMaxNormalize()
    res_norm = RescaleNormalize(clip_range=(2, 98))

    s_img = s_norm(x=img)
    assert s_img.min() >= 0 and s_img.max() <= 1
    assert s_img.dtype == np.float32

    mm_img = mm_norm(x=img)
    assert mm_img.min() >= -1 and mm_img.max() <= 1
    assert mm_img.dtype == np.float32

    res_img = res_norm(x=img)
    assert res_img.min() >= -1 and res_img.max() <= 255
    assert res_img.dtype == np.uint8


class TestDataSetBuilder2D3D:
    def test_data_augmentation3d(self):
        img = np.random.rand(64, 64, 64)
        img = img.astype(np.float32)

        img_proc, mask = preprocess(image=img, mask=img, transformation=True, size=64)
        assert img_proc.shape == (1, 64, 64, 64)
        assert mask.shape == (1, 64, 64, 64)
        assert img_proc.dtype == np.float32
        assert np.all(img_proc == mask)
        assert img_proc.min() >= -1 and img_proc.max() <= 1
        assert mask.min() >= 0 and mask.max() <= 1

    def test_data_augmentation2d(self):
        img = np.random.rand(64, 64)
        img = img.astype(np.float32)

        img_proc, mask = preprocess(image=img, mask=img, transformation=True, size=64)
        assert img_proc.shape == (1, 64, 64)
        assert mask.shape == (1, 64, 64)
        assert img_proc.dtype == np.float32
        assert np.all(img_proc == mask)
        assert img_proc.min() >= -1 and img_proc.max() <= 1
        assert mask.min() >= 0 and mask.max() <= 1



# #####################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
# #####################################################################

###
# The main function to run the any analysis for fiber
# Working:
# - Get data set work on one dataset at a time, so require file not folder
# - flags to run globla analysis
#     - Lenght
#     - Curv/tourrositi
#     - fiber classification (require chromosome and poles)
#     - Global interaction map

# The rest need napari ...
###



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
from os import mkdir

import click
import numpy as np
import os

from tardis_em_analysis.mt_classification.mt_classes import MicrotubuleClassifier
from tardis_em.utils.export_data import NumpyToAmira
from tardis_em._version import version


@click.command()
@click.option(
    "-s",
    "--path_surface",
    type=str,
    help="File directory to .surf Amira file.",
    show_default=True,
)
@click.option(
    "-f",
    "--path_filaments",
    type=str,
    help="File directory to .am spatial graph Amira file with microtubules.",
    show_default=True,
)
@click.option(
    "-p",
    "--path_poles",
    type=str,
    help="File directory to .am spatial graph Amira file with poles.",
    show_default=True,
)
@click.option(
    "-px",
    "--pixel_size",
    type=float,
    default=None,
    help="Pixel size value.",
    show_default=True,
)
@click.option(
    "-gap",
    "--gap_size",
    type=int,
    default=100,
    help="Gap size to fill out. Assuming very uneven surface to help the tool estimate when MT"
    "enter and exit the volume, the small gaps can be marge on the microtubule.",
    show_default=True,
)
@click.option(
    "-dist",
    "--kmts_to_surface_distance",
    type=int,
    default=1000,
    help="MT (+) end distance to the NN surface which allow it be be assigned as KMTs..",
    show_default=True,
)
@click.option(
    "-v",
    "--visualize",
    type=bool,
    default=False,
    help="MT (+) end distance to the NN surface which allow it be be assigned as KMTs..",
    show_default=True,
)
@click.version_option(version=version)
def main(
    path_surface: str,
    path_filaments: str,
    path_poles: str,
    pixel_size: float,
    gap_size: int,
    kmts_to_surface_distance: int,
    visualize: bool,
):
    dir_ = os.path.dirname(path_surface)
    out_ = os.path.join(dir_, "analysis")
    out_name = os.path.split(path_filaments)[-1][:-3] + "_classes.am"
    out_name = os.path.join(dir_, "analysis", out_name)

    if not os.path.isdir(out_):
        mkdir(out_)

    classifier = MicrotubuleClassifier(
        surfaces=path_surface,
        filaments=path_filaments,
        poles=path_poles,
        pixel_size=pixel_size,
        gaps_size=gap_size,
        kmt_dist_to_surf=kmts_to_surface_distance,
    )

    classifier.classified_MTs()

    classes = classifier.get_classified_fibers()

    am = NumpyToAmira(
        as_point_cloud=False,
        header=[
            f"Classification Analysis done with TARDIS_em v{version}",
            "Segmentation results:",
            f"KMTs: {len(classes[0][0]) + len(classes[0][1]) + len(classes[1][0]) + len(classes[1][1])}",
            f"Mid-MTs: {len(classes[2])}",
            f"Interdigitating-MTs: {len(classes[3])}",
            f"Bridging-MTs: {len(classes[4])}",
            f"SMTs: {len(classes[5])}",
        ],
    )

    am.export_amira(
        file_dir=out_name,
        coords=[
            classes[5],
            classes[0][0],
            classes[0][1],
            classes[1][0],
            classes[1][1],
            classes[2],
            classes[3],
            classes[4],
        ],
        labels=[
            "SMTs",
            "KMTs_pole1_Inside",
            "KMTs_pole1_Outside",
            "KMTs_pole2_Inside",
            "KMTs_pole2_Outside",
            "Mid-MTs",
            "Interdigitating-MTs",
            "Bridging-MTs",
        ],
    )

    if visualize:
        try:
            from tardis_em.utils.visualize_pc import (
                VisualizeFilaments,
                VisualizeSurface,
                VisualizeCompose,
            )

            vertices, triangles = classifier.get_vertices(simplify=128)
            vis_kmt = VisualizeFilaments(
                np.vstack((np.vstack(classes[0]), np.vstack(classes[1]))),
                False,
                filament_color="red",
                return_b=True,
            )
            vis_mid_mt = VisualizeFilaments(
                classes[2], False, filament_color="magenta", return_b=True
            )
            vis_brg_mt = VisualizeFilaments(
                classes[3], False, filament_color="white", return_b=True
            )
            vis_int_mt = VisualizeFilaments(
                classes[4], False, filament_color="orange", return_b=True
            )
            vis_smt = VisualizeFilaments(
                classes[5], False, filament_color="yellow", return_b=True
            )
            surf = VisualizeSurface(vertices, triangles, False, return_b=True)

            VisualizeCompose(
                False,
                vis_smt=vis_smt,
                vis_kmt=vis_kmt,
                vis_mid_mt=vis_mid_mt,
                vis_int_mt=vis_int_mt,
                vis_brg_mt=vis_brg_mt,
                surf=surf,
            )
        except ModuleNotFoundError:
            pass


if __name__ == "__main__":
    main()
